#include <stdio.h>
#include "expr.c"

enum { EXP_ADD=XPP_LAST, EXP_SUB, EXP_DIV, EXP_MOD, EXP_MUL,
       EXP_NEG, EXP_PLUS, EXP_PAREN_R, EXP_FCALL,
       EXP_PAREN_C, EXP_PAREN_S, EXP_ARRAY, EXP_SET, EXP_COMMA , EXP_ABS,
       EXP_QUOTE };

#define TEST 2

void prt_node(expr_t *E)
{
  static int tab;
  int i;
  for(i=0;i<tab;i++) printf(" ");
  switch(E->op)
  {
  case XPP_INT: printf("int(%s)", E->data); break;
  case XPP_FLO: printf("flo(%s)", E->data); break;
  case XPP_IDENT: printf("ident(%s)", E->data); break;
  case EXP_ADD: printf("add"); break;
  case EXP_SUB: printf("sub"); break;
  case EXP_MUL: printf("mul"); break;
  case EXP_DIV: printf("div"); break;
  case EXP_MOD: printf("mod"); break;
  case EXP_NEG: printf("neg"); break;
  case EXP_PLUS: printf("plus"); break;
  case EXP_PAREN_R: printf("parenr"); break;
  case EXP_PAREN_C: printf("parenc"); break;
  case EXP_PAREN_S: printf("parens"); break;
  case EXP_FCALL: printf("fcall"); break;
  case EXP_ARRAY: printf("array"); break;
  case EXP_SET: printf("set"); break;
  case EXP_COMMA: printf("comma"); break;
  case EXP_ABS: printf("abs"); break;
  case EXP_QUOTE: printf("quote(%s)", E->data); break;
  default: printf("--unknown %d--", E->op);
  }
  printf("\n");
  tab+= 3;
  if (E->left) prt_node(E->left);
  if (E->right) prt_node(E->right);
  tab-= 3;
}
       

int main(int argc,char **argv)
{
  xpp_t *xpp;
  expr_t *E;

  xpp= xpp_new();
  xpp_binary_operator(xpp, "+", EXP_ADD, 50, 'L');
  xpp_binary_operator(xpp, "-", EXP_SUB, 50, 'L');
  xpp_binary_operator(xpp, "*", EXP_MUL, 60, 'L');
  xpp_binary_operator(xpp, "/", EXP_DIV, 60, 'L');
  xpp_binary_operator(xpp, "%", EXP_MOD, 60, 'L');
  xpp_prefix_operator(xpp, "-", EXP_NEG);
  xpp_prefix_operator(xpp, "+", EXP_PLUS);
  xpp_binary_operator(xpp, "=", EXP_SET, 30, 'R');
  xpp_binary_operator(xpp, ",", EXP_COMMA, 20, 'R');

  xpp_parenthesis(xpp, "(", ")", EXP_PAREN_R);
  xpp_parenthesis(xpp, "[", "]", EXP_PAREN_S);
  xpp_parenthesis(xpp, "{", "}", EXP_PAREN_C);
  xpp_parenthesis(xpp, "|(", ")|", EXP_ABS);
  xpp_postfix_parenthesis(xpp, "(", ")", EXP_FCALL);
  xpp_postfix_parenthesis(xpp, "[", "]", EXP_ARRAY);
  xpp_quote(xpp,"\"", "\"", '\\', EXP_QUOTE);

#if TEST==1
  xpp->inp= argv[1];
  do
  {
    lex(xpp);
    printf("tok= %d text= %s (op= %d)\n",
           xpp->token, xpp->toktext, xpp->tokoper);
  } while(xpp->token!=TOK_EOF && xpp->token!=TOK_ERROR);
#endif

#if TEST==2
  E= xpp_run(xpp, argv[1]);
  if (E->op==XPP_ERROR)
  {
    printf("error: %s\n", E->data);
    printf("%s\n", argv[1]);
    int i;
    for(i=0;i<E->pos;i++) printf(" ");
    printf("^\n");
  }
  else
    prt_node(E);
  xpp_free_expr(E);
  xpp_free(xpp);
#endif

  return 0;
}
