#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include "strbuf.h"
#include "expr.h"
#include "widget.h"
#include "bytecode.h"
#include "n-eval.h"
#include "strmap.h"

typedef struct { char *str; int iv; } strintpair_t;

static strmap_t *symbols;

typedef struct
{
  int type;    // 'W'= widget coord, 'V'= local var, 'G'= global var
  int visit;           // flag to exclude infinite recursion

  expr_t *expr;
  bc_buffer_t *access; // code to access this thing. this is non-null
                       // if the computation has been done successfully.
  Uwidget_t *widget;
  int wnum;
  int coord_idx; // 0..5 for widget coord

  Uvar_t *var;

  int localnum;
} value_t;

typedef struct
{
  Uwidget_t *widget;
  char *ident;
  int wnum; // within the screen
  value_t *x,*y,*w,*h,*bot,*rig;
} widget_eval_t;


static int n_values;
static value_t **values;
static int n_locals;
static int n_widgets;
static widget_eval_t *wevals;

void evaldis_coords(Uscreen_t *S,int sw,int sh);
char *eval_screen_dim(Uscreen_t *S);

static int coord_accessor(int csel)
{
  switch(csel)
  {
  case wTOP: return GUII_WLAY_Y;
  case wLEF: return GUII_WLAY_X;
  case wWID: return GUII_WLAY_W;
  case wHEI: return GUII_WLAY_H;
  }
  return GUII_WLAY_Y;
}




static bc_buffer_t* bc_buffer_new(int size)
{
  bc_buffer_t *B;
  B= malloc(sizeof(*B));
  B->size= size;
  B->len= 0;
  B->data= malloc(size);
  return B;
}




static void bc_buffer_put(bc_buffer_t *buf, unsigned char V)
{
  if (buf->len==buf->size)
  {
    buf->size += buf->size/2 + 20;
    buf->data= realloc(buf->data, buf->size);
  }
  buf->data[buf->len++]= V;
}

#define put_byte   bc_buffer_put
#define put_code   put_byte

static int bc_buffer_tell(bc_buffer_t *buf)
{
  return buf->len;
}

static void bc_buffer_rewind(bc_buffer_t *buf, int pos)
{
  buf->len= pos;
}

static void bc_buffer_append(bc_buffer_t *dst, bc_buffer_t *src)
{
  if (dst->len+src->len>dst->size)
  {
     dst->size= dst->len+src->len+dst->size/4;
     dst->data= realloc(dst->data, dst->size);
  }
  memcpy(dst->data+dst->len, src->data, src->len);
  dst->len+= src->len;
}



static int alloc_local_var()
{
  return n_locals++;
}


static expr_t *make_widget_expression(Uwidget_t *W, char *f1, int op, char *f2)
{
  char buf[200];
  expr_t *e;
  sprintf(buf,"%s.%s %c %s.%s",W->ident,f1, op, W->ident, f2);
  e= xpp_run(xpp, buf);
  if (!e || e->op==XPP_ERROR) e= xpp_run(xpp, "0");
  return e;
}

static value_t* emit_coord_value
  (Uwidget_t *W, int wnum, int coord_idx, 
   expr_t *e, char *altF1, int op, char *altF2)
{
  value_t *V;
  if (!e) e= make_widget_expression(W, altF1, op, altF2);
  V= calloc(1,sizeof(*V));
  V->expr= e;
  V->type= 'W';
  V->widget= W;
  V->wnum= wnum;
  V->coord_idx= coord_idx;
  return values[n_values++]= V;
}

static value_t* emit_coord_value_as_local
  (Uwidget_t *W, int wnum, int coord_idx,
   expr_t *e, char *altF1, int op, char *altF2)
{
  value_t *V;
  if (!e) e= make_widget_expression(W, altF1, op, altF2);
  V= calloc(1,sizeof(*V));
  V->expr= e;
  V->type= 'V';
  V->localnum= alloc_local_var();
  values[n_values++]= V;

  V->wnum= wnum;
  V->widget= W;
  V->coord_idx= coord_idx;

  return  V;
}
   

static void widget_to_weval(Uwidget_t *W,int wnum,widget_eval_t *L)
{
  L->x= emit_coord_value(W, wnum, wLEF, 
		      W->pos.E[wLEF],  "right", '-', "width" );

  L->y= emit_coord_value(W, wnum, wTOP, 
		      W->pos.E[wTOP],  "bottom", '-', "height" );

  L->w= emit_coord_value(W, wnum, wWID,
		      W->pos.E[wWID],  "right", '-', "left" );
  L->h= emit_coord_value(W, wnum, wHEI,
		      W->pos.E[wHEI],  "bottom", '-',  "top" );

  L->bot= emit_coord_value_as_local(W, wnum, wBOT,
                      W->pos.E[wBOT], "top", '+', "height" );
  L->rig= emit_coord_value_as_local(W, wnum, wRIG,
                      W->pos.E[wRIG], "left", '+', "width" );
  L->wnum= wnum;
  L->widget= W;
  L->ident= W->ident;
}




static void make_wevals(Uscreen_t *S)
{
  Uwidget_t *W;
  int N;
  for(W=S->widgets,N=0;W;W=W->next,N++)
	  ;
  n_widgets= N;
  wevals= malloc(sizeof(wevals[0])*N);
  for(W=S->widgets,N=0;W;W=W->next,N++)
     widget_to_weval(W, N, wevals+N);
}

static void emit_var_values(Uscreen_t *S)
{
  Uvar_t *var;
  value_t *V;
  for(var=S->vars;var;var=var->next)
  {
    V= calloc(1,sizeof(*V));
    V->type= 'V';
    V->var= var;
    V->expr= var->expr;
    V->localnum= alloc_local_var();
    values[n_values++]= V;
  }
}

static int count_vars(Uscreen_t *S)
{
  int n; Uvar_t *V;
  for(V=S->vars,n=0;V;V=V->next,n++) ;
  return n;
}

static int count_widgets(Uscreen_t *S)
{
  int n; Uwidget_t *W;
  for(W=S->widgets,n=0;W;W=W->next, n++) ;
  return n;
}


static bc_buffer_t *cbuf;

static value_t *error_value;
static char *error_message;

static int err1arg(char *msg, char *arg)
{
  char *R;
  R= malloc(strlen(msg)+strlen(arg)+1);
  sprintf(R, "%s%s",msg,arg);
  error_message= R;
  return 1;
}

static widget_eval_t *find_weval(char *name)
{
  int i;
  for(i=0;i<n_widgets;i++) 
     if (!strcmp(wevals[i].ident,name)) return wevals+i;
  return NULL;
}

static int compute_value(value_t *V);

static int compute_dot_expr(expr_t *L, expr_t *R)

{
  widget_eval_t *WE;
  value_t *V;
  if (L->op!=XPP_IDENT || R->op!=XPP_IDENT) return 1;
  if (!(WE=find_weval(L->data))) return err1arg("unknown widget ", L->data);
  
  if (!strcmp(R->data,"left")) V= WE->x;
  else if (!strcmp(R->data,"right")) V= WE->rig;
  else if (!strcmp(R->data,"top")) V= WE->y;
  else if (!strcmp(R->data,"bottom")) V= WE->bot;
  else if (!strcmp(R->data, "width")) V= WE->w;
  else if (!strcmp(R->data, "height")) V= WE->h;
  else return err1arg("unknown field ",R->data);

  if (compute_value(V)) return 1;
  bc_buffer_append(cbuf, V->access);
  return 0;
}

static value_t *find_var_value(char *name)
{
  int i;
  value_t *V;
  for(i=0;i<n_values;i++)
  {
    V= values[i];
    if (V->type=='V' && V->var && !strcmp(V->var->ident,name)) return V;
  }
  return NULL;
}

static int compute_ident(char *name)
{
  value_t *V;
  if (!(V= find_var_value(name))) return err1arg("unknown variable ",name);
  if (compute_value(V)) return 1; 
  bc_buffer_append(cbuf, V->access);
  return 0;
}

static void put_int(bc_buffer_t *B, uint32_t V)
{
  if (V&(0xff<<24))
  {
    put_byte(B, GUII_C4);
    put_byte(B, V>>24);
    put_byte(B, V>>16);
    put_byte(B, V>>8);
    put_byte(B, V);
  }
  else if (V&(0xff<<16))
  {
    put_byte(B, GUII_C3);
    put_byte(B, V>>16);
    put_byte(B, V>>8);
    put_byte(B, V);
  }
  else if (V&(0xff<<8))
  {
    put_byte(B, GUII_C2);
    put_byte(B, V>>8);
    put_byte(B, V);
  }
  else if ( V & 0x80 )
  {
    put_byte(B, GUII_C1);
    put_byte(B, V);
  }
  else put_byte(B, 0x80 | V );
}

static void put_flo(bc_buffer_t *B, float F)
{
  union { float F; uint32_t I; } D;
  uint32_t V;
  D.F= F;
  V= D.I;

  put_byte(B, GUII_CF);
  put_byte(B, V>>24);
  put_byte(B, V>>16);
  put_byte(B, V>>8);
  put_byte(B, V);

}

static int compute_expr(expr_t *E);

static int compute_binop(expr_t *E, unsigned char op)
{
  if (compute_expr(E->left)) return 1;
  if (compute_expr(E->right)) return 1;
  put_code(cbuf,op);
  return 0;
}

static int compute_expr(expr_t *E)
{
  int i;
  switch(E->op)
  {
   case EXP_ADD: return compute_binop(E, GUII_ADD);
   case EXP_SUB: return compute_binop(E, GUII_SUB);
   case EXP_MUL: return compute_binop(E, GUII_MUL);
   case EXP_DIV: return compute_binop(E, GUII_DIV);
   case EXP_MOD: return compute_binop(E, GUII_MOD);
   case EXP_NEG:
     put_int(cbuf, 0); if (compute_expr(E->right)) return 1;
     put_code(cbuf, GUII_SUB);
     return 0;
   case EXP_PLUS: return compute_expr(E->right);
   case EXP_PAREN_R: return compute_expr(E->left);
   case XPP_FLO: put_flo(cbuf, strtof(E->data, NULL)); return 0;
   case XPP_INT: put_int(cbuf, strtol(E->data,NULL,0)); return 0;
   case XPP_ERROR: return 1;
   case XPP_IDENT:
     return compute_ident(E->data);
   case EXP_DOT:
     return compute_dot_expr(E->left, E->right);
  }
  return 1;
}


static int compute_value(value_t *V)
{
  int R;
  if (V->access) return 0; // already known
  if (V->visit) 
  {
    error_value= V;
    error_message= "A circular computation is found";
    return 1;
  }
  V->visit= 1;
  R= compute_expr(V->expr);
  V->visit= 0;
  if (R) 
  {
    if (!error_value)  error_value= V; 
    if (!error_message)  error_message= "Can not compute"; 
    return 1;
  }
  V->access= bc_buffer_new(50);
  if (V->type=='W')
  {
     put_int(cbuf, V->wnum);
     put_code(cbuf, coord_accessor(V->coord_idx));
     put_code(cbuf, GUII_STOREINT);

     put_int(V->access, V->wnum);
     put_code(V->access, coord_accessor(V->coord_idx));
     put_code(V->access, GUII_LOADINT);
  }
  else
  {
     put_int(cbuf, V->localnum);
     put_code(cbuf, GUII_LOCAL);
     put_code(cbuf, GUII_STOREINT);

     put_int(V->access, V->localnum);
     put_code(V->access, GUII_LOCAL);
     put_code(V->access, GUII_LOADINT);
  }
  // FIXME: handle globals
  return 0;
}

static int compute_widget_coords(widget_eval_t *L)
{
  if (compute_value(L->x) 
     || compute_value(L->y) 
     || compute_value(L->w)
     || compute_value(L->h))
   return 1;
  return 0;  
}


void bc_buffer_free(bc_buffer_t *buf)
{
  free(buf->data);
  free(buf);
}


static void emit_predef_locals()
{
  value_t *V;
  int i;
  char *name; int localnum;
  static const strintpair_t predefs[]=
  {
    { "WIDTH", GUILV_WIDTH },
    { "HEIGHT", GUILV_HEIGHT },
    { "font1height", GUILV_FONT1HEIGHT },
    { "font2height", GUILV_FONT2HEIGHT },
    { "font3height", GUILV_FONT3HEIGHT },
    { "font4height", GUILV_FONT4HEIGHT },
    { "font5height", GUILV_FONT5HEIGHT },
    { "gui_scrollbar_width", GUILV_GUI_SCROLLBAR_WIDTH },
    { "gui_scrollbar_height", GUILV_GUI_SCROLLBAR_HEIGHT },
    { "gui_sysbar_height", GUILV_GUI_SYSBAR_HEIGHT },
    { NULL, 0 }
  };
  for(i=0;predefs[i].str;i++)
  {
    name= predefs[i].str; localnum= predefs[i].iv;

    Uvar_t *var;
    var= malloc(sizeof(*var));
    var->ident= strdup(name);
    value_t *V;
    V= calloc(1,sizeof(*V));
    V->type= 'V';
    V->var= var;
    V->localnum= localnum;
    V->access= bc_buffer_new(50);
    put_int(V->access, localnum);
    put_code(V->access, GUII_LOCAL);
    put_code(V->access, GUII_LOADINT);
    values[n_values++]= V;
  }
}

typedef struct
{
  strbuf_t *B;
  int linelen;
} Gstrbuf_t;

static Gstrbuf_t *Gstrbuf_new()
{
  Gstrbuf_t *G;
  G= malloc(sizeof(*G));
  G->B= strbuf_init();
  G->linelen= 0;
  return G;
}

static void GputS(Gstrbuf_t *G, char *K)
{
  int L;
  L= strlen(K);
  if (G->linelen+L>72)
  {
    strbuf_puts_l(G->B, "\n", 1);
    G->linelen= 0;
  }
  strbuf_puts_l(G->B, K, L);
  G->linelen+= L;
}

static void GputB(Gstrbuf_t *G,unsigned char X)
{
  static char buf[20];
  sprintf(buf,"%d,",X);
  GputS(G,buf);
}

static void GputI(Gstrbuf_t *G,int X)
{
  static char buf[20];
  sprintf(buf,"%d,",X);
  GputS(G,buf);
}

static char *Gfree(Gstrbuf_t *G)
{
  char *R;
  R= strbuf_collect(G->B);
  free(G);  
  return R;
}


char *bc_buffer_to_string(bc_buffer_t *cb, int n_loc,int only_code)
{
  int i;
  Gstrbuf_t *G;
  G= Gstrbuf_new();
  if (!only_code)
  {
    GputS(G, " .n_locals= ");
    GputI(G, n_loc);
    GputS(G, " .length= ");
    GputI(G, cb->len);
    GputS(G, " .data= (const unsigned char[]) {");
  }
  for(i=0;i<cb->len;i++) GputB(G, cb->data[i]);
  if (!only_code) GputS(G, "}");
  return Gfree(G);
}


bc_buffer_t *compile_widget_layout(Uscreen_t *S,int *n_loc)
{
  int nu;
  int i;
  symbols= strmap_new();

  nu= count_vars(S) + count_widgets(S) * 6 + n_predef_locals;
  n_values= 0;
  values= calloc(nu, sizeof(values[0]));
  emit_predef_locals();
  n_locals= n_predef_locals;
  emit_var_values(S);
  make_wevals(S);
  cbuf= bc_buffer_new(256);
  
  for(i=0;i<n_widgets;i++)
    if (compute_widget_coords(wevals+i)) goto fail;

back_from_fail:
  //
  // there is some memory leak due to expressions that were created
  // by us (ie. u.right= u.left+u.width etc). I'm going to ignore them.
  //
  for(i=0;i<n_values;i++) free(values[i]);
  free(values);
  free(wevals);
  *n_loc= n_locals;

  return cbuf;

fail:
  bc_buffer_free(cbuf);
  cbuf= NULL;
  goto back_from_fail;
}

void print_bytecode(char *fname, bc_buffer_t *B)
{
  FILE *f;
  int i;
  char *K;
  f= fopen(fname, "w");
  if (!f) return ;
  K= bc_buffer_to_string(B, n_locals,0);
  fputs(K,f);
  fclose(f);
  free(K);
}
