#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include "bytecode.h"

#define uint32_t u_int32_t

static uint32_t getI(unsigned char *data,int siz)
{
  int i;
  uint32_t V,D;
  V= 0;
  for(i=0;i<siz;i++)
  {
    V <<= 8;
    D= data[i];
    V |= D;
  }
  return V;
}

static float getF(unsigned char *data)
{
  union { float F; uint32_t I; } D;
  D.I= getI(data,4);
  return D.F;
}

static void print_bytes(unsigned char *data, int len)
{
  int i;
  for(i=0;i<len;i++) printf("%02x ",data[i]);
  for(;i<6;i++)      printf("   ");
  printf(" | ");
}


static void disas(unsigned char *data,int len)
{
  int pc;
  int i;
  unsigned char C;
  char *P;
  uint32_t VI;
  float VF;

  pc= 0;
  while(pc<len)
  {
    printf("%4d: ", pc);
    C= data[pc];
    if (C&0x80)
    {
      print_bytes(data+pc, 1);
      printf(" CONST_INT(%d)\n", C&0x7f);
      pc++;
      continue;
    }
    switch(C)
    {
case GUII_C4          : print_bytes(data+pc,5); VI= getI(data+pc+1, 4);
                        printf("CONST_INT(0x%x)\n", VI); pc+=5; continue;
case GUII_C3          : print_bytes(data+pc,4); VI= getI(data+pc+1, 3);
                        printf("CONST_INT(0x%x)\n", VI); pc+=4; continue;
case GUII_C2          : print_bytes(data+pc,3); VI= getI(data+pc+1, 2);
                        printf("CONST_INT(%d)\n", VI); pc+= 3; continue;
case GUII_C1          : print_bytes(data+pc,2); VI= getI(data+pc+1, 1); 
                        printf("CONST_INT(%d)\n", VI); pc+= 2; continue;
case GUII_CF          : print_bytes(data+pc,5); VF= getF(data+pc+1);
                        printf("CONST_FLO(%f)\n", VF); pc+= 5; continue;
default: print_bytes(data+pc, 1); 
    }
    switch(C)
    {
case GUII_ADD         : P= "ADD"; break;
case GUII_SUB         : P= "SUB"; break;
case GUII_MUL         : P= "MUL"; break;
case GUII_DIV         : P= "DIV"; break;
case GUII_MOD         : P= "MOD"; break;
case GUII_STOREINT    : P= "STOREINT"; break;
case GUII_LOADINT     : P= "LOADINT"; break;
case GUII_LOCAL       : P= "LOCAL"; break;
case GUII_WLAY_X      : P= "WLAY_X"; break;
case GUII_WLAY_Y      : P= "WLAY_Y"; break;
case GUII_WLAY_W      : P= "WLAY_W"; break;
case GUII_WLAY_H      : P= "WLAY_H"; break;
default               : P= "-unknown-"; 
    }
    printf("%s\n", P);
    pc++;
  }
}

static int getint(FILE *F)
{
  char k;
  int V;
  V= 0;

again:
  if (fread(&k,1,1,F)<1) return V;
  if (k<'0' || k>'9') goto again;

  V= k-'0';
digit:
  if (fread(&k,1,1,F)<1) return V;
  if (k<'0' || k>'9') return V;
  V= V*10 + k - '0';
  goto digit;
}

int main(int argc,char **argv)
{
  char *fn;
  int len,locs;
   int i;
  FILE *F;
  unsigned char *data;
  if (argc!=2) exit(fprintf(stderr,"usage: %s filename\n",argv[0]));
  fn= argv[1];
  F= fopen(fn,"rb");
  if (!F) exit(fprintf(stderr,"%s: can't open %s\n",argv[0],fn));
  locs= getint(F);
  len= getint(F);
  data= malloc(len);
  for(i=0;i<len;i++) data[i]= getint(F);
  fclose(F);
  printf("=== N_LOCALS: %d LENGTH: %d ===\n", locs, len);
  disas(data,len);
  return 0;
}

