/***********************************************************************
 *                                                                     *
 *                           Widget Editors                            *
 *                                                                     *
 ***********************************************************************
   Specialized widget editors live here. These are called ew_WIDGET().
 ***********************************************************************/
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdint.h>
#include "ptrbuf.h"
#include "strbuf.h"
#include "expr.h"
#include "widget.h"
#include "gfx.h"

static int ewfmode;
   // reading is done from the screen
   // writing is done to the screen
static void ewf_read() { ewfmode= 'R'; }
static void ewf_write() { ewfmode= 'W'; }

static void remove_field(Uwidget_t *W,char *fi)
{
  field_t **T,*f;
  for(T=&W->fields,f=W->fields;f;)
  {
     if (!strcmp(f->name,fi))
     {
        *T= f->next;
        free(f);
        return ;
     }
     T= &f->next;
     f= f->next;
  }
}

static field_t *find_field(Uwidget_t *W,char *fi)
{
  field_t *f;
  for(f=W->fields;f;f=f->next)
     if (!strcmp(f->name,fi))
       return f;
  return NULL;
}
static field_t *find_klass_field(Uklass_t *K,char *fi)
{
  int i;
  for(i=0;i<K->nfields;i++)
     if (!strcmp(K->fields[i].name,fi))
       return K->fields+i;
  return NULL;
}

static void set_field(Uwidget_t *W,char *fi,void *V)
{
  field_t *f,*kf;
  f= find_field(W,fi);
  if (!f)
  {
    kf= find_klass_field(W->klass,fi);
    if (!kf) return ;
    f= calloc(1,sizeof(*f));
    f->type= kf->type;
    f->name= kf->name;
    f->next= W->fields;
    W->fields= f;
  }
  f->data= V;
}

static void ewf_text(int wno, Uwidget_t *W,char *fi)
{
  if (ewfmode=='R')
  {
    char *s;
    s= gfx_get_text(wno);
    if (!s || !s[0])
      remove_field(W,fi);
    else
      set_field(W, fi, s);
  }
  else
  {
    field_t *f;
    f= find_field(W,fi);
    if (!f) f= find_klass_field(W->klass,fi);
    if (!f || !f->data)
      gfx_set_text(wno,"");
    else
      gfx_set_text(wno,f->data);
  }
}

static int ewloop(int ok,int canc)
{
  gfx_event_t event;
  while(1)
   if (gfx_event(NULL,&event))
   {
    if (event.type==canc || event.type==GFXEV_QUIT) return 0;
    if (event.type==ok) return 1;
   }
}

static void ewf_enum(int wno, Uwidget_t *W, char *fi, char *edef)
{
  if (ewfmode=='R')
  {
    char *B,*D,*E,*N;
    int sel,count;
    sel= gfx_get_selected(wno);
    if (sel==-1) 
    {
      remove_field(W,fi);
      return ;
    }

    D= edef= strdup(edef);
    count= 0;
    while(D)
    {
       B= strchr(D, '@');
       if (!B) break;
       E= strchr(B,'#');
       B[0]= 0;
       if (E) { E[0]= 0; N= E+1; } else N= E;
       if (sel==count) { set_field(W,fi,strdup(D)); break; }
       count++;
       D= N;
    }
    free(edef);
  }
  else
  {
    char *B,*D,*E,*N;
    int sel,count;
    field_t *f;
    f= find_field(W,fi);
    if (!f) f= find_klass_field(W->klass,fi);
    sel= -1; count= 0;

    D= edef= strdup(edef);
    while(D)
    {
       B= strchr(D, '@');
       if (!B) break;
       E= strchr(B,'#');
       B[0]= 0;
       if (E) { E[0]= 0; N= E+1; } else N= E;
       if (f && f->data && !strcmp(f->data,D)) sel= count;
       gfx_insert(wno,-1,B+1,NULL,NULL);
       count++;
       D= N;
    }
    if (sel!=-1) gfx_set_selected(wno,sel);
    free(edef);
  }
}

static void ewf_bool(int wno, Uwidget_t *W, char *fi)
{
  if (ewfmode=='R')
  {
    char *v;
    v= gfx_get_state(wno) ? "1" : "0";
    set_field(W,fi, strdup(v));
  }
  else
  {
    field_t *f;
    int v;

    f= find_field(W,fi);
    if (!f) f= find_klass_field(W->klass,fi);
    if (!f) return ;
    if (f->data && atoi(f->data)) v= 1;
                             else v= 0;
    gfx_set_state(wno, v);
  }
}

static void ew_label_attribs(Uwidget_t *W)
{
  ewf_text(EWlabTEXT,W,"text");
  ewf_text(EWlabFG, W, "fg");
  ewf_text(EWlabBG, W, "bg");
  ewf_enum(EWlabJUST, W, "just", "L@Left#C@Center#R@Right");
  ewf_enum(EWlabFONT, W, "font", "1@1#2@2#3@3");
}

void ew_label(Uwidget_t *W)
{
  gfx_SCRewLABEL();

  ewf_write();
  ew_label_attribs(W);
  if (ewloop(EWlabOK,EWlabCANCEL))
  {
    ewf_read();  
    ew_label_attribs(W);
  }
  gfx_leave_screen();
}

static void ew_filelist_attribs(Uwidget_t *W)
{
  ewf_bool(EWfilelistHIDDEN,W, "hidden");
  ewf_bool(EWfilelistTYPES,W, "types");
}

void ew_filelist(Uwidget_t *W)
{
  gfx_SCRewFILELIST();
  ewf_write();
  ew_filelist_attribs(W);
  if (ewloop(EWfilelistOK, EWfilelistCANCEL))
  {
    ewf_read();
    ew_filelist_attribs(W);
  }
  gfx_leave_screen();
}

static void ew_progressbar_attribs(Uwidget_t *W)
{
  ewf_text(EWprogFG, W, "fg");
  ewf_text(EWprogBG, W, "bg");
  ewf_text(EWprogFILL, W, "fill");
  ewf_enum(EWprogFONT, W, "font", "1@1#2@2#3@3");
}


void ew_progressbar(Uwidget_t *W)
{
  gfx_SCRewPROGRESS();

  ewf_write();
  ew_progressbar_attribs(W);
  if (ewloop(EWprogOK,EWprogCANCEL))
  {
    ewf_read();  
    ew_progressbar_attribs(W);
  }
  gfx_leave_screen();
}


static void ew_canvas_attribs(Uwidget_t *W)
{
  ewf_bool(EWcanHORSCROLL, W, "horscroll");
  ewf_bool(EWcanVERSCROLL, W, "verscroll");
}

void ew_canvas(Uwidget_t *W)
{
  gfx_SCRewCANVAS();
  ewf_write();
  ew_canvas_attribs(W);
  if (ewloop(EWcanOK,EWcanCANCEL))
  {
    ewf_read();  
    ew_canvas_attribs(W);
  }
  gfx_leave_screen();
}


void ew_panel(Uwidget_t *W)
{
  int i,j;
  int q;
  char buf[200];
  gfx_event_t event;
  int sel_idx; 
  Uscreen_t *S;

  gfx_SCRewPANEL();

  for(i=0;i<W->n_subs;i++)
    gfx_insert(EWpanLIST, -1, W->subs[i]->ident, W->subs[i], NULL);

  q= 0;
  sel_idx= -1;
  while(!q) if (gfx_event(NULL,&event)) switch(event.type)
  {
  case GFXEV_QUIT:
  case EWpanOK: q= 1; break;
  case EWpanNEW:
    S= calloc(1,sizeof(*S));
    for(j=0;j<1000;j++)
    {
      sprintf(buf,"%s.%d", W->ident, j);
      for(i=0;i<W->n_subs;i++)
          if (!strcmp(W->subs[i]->ident,buf)) break;
      if (i==W->n_subs) break;
    }
    S->type= 1;
    S->ident= strdup(buf);
    if (W->n_subs)
      W->subs= realloc(W->subs, (W->n_subs+1)*sizeof(W->subs[0]));
    else
      W->subs= malloc((W->n_subs+1)*sizeof(W->subs[0]));
    W->subs[W->n_subs]= S;
    W->n_subs++;
    gfx_insert(EWpanLIST, -1, S->ident, S, NULL);
    break;
  case EWpanLIST:
    sel_idx= event.key;
    break;
  case EWpanDEL:
    if (sel_idx==-1) break;
    S= W->subs[sel_idx];
    if (S->widgets && gfx_ask("Confirm Screen Deletion","!OK!Cancel",
                              "Selected screen %s is not empty.\n"
                              "Proceed with deletion?", S->ident))
        break;
    for(i=sel_idx;i<W->n_subs-1;i++)
      W->subs[i]= W->subs[i+1];
    gfx_remove(EWpanLIST, sel_idx, NULL);
    sel_idx= -1;
    break;
  }
  gfx_leave_screen();
}

void ew_button_attribs(Uwidget_t *W)
{
  ewf_text(EWbuttonTEXT,W,"text");
  ewf_enum(EWbuttonICONPOS, W, "iconpos", "T@Top#B@Bottom#L@Left#R@Right");
  ewf_text(EWbuttonICON,W,"icon");
}

static void ew_button_generic(Uwidget_t *W,char *title)
{
  char *ic;
  gfx_event_t event;
  gfx_SCRewBUTTON();

  gfx_set_text(EWbuttonTITLE, title);

  ewf_write();
  ew_button_attribs(W);
  while(1)
   if (gfx_event(NULL,&event))
   switch(event.type)
   {
   case EWbuttonOK:
      ewf_read();  
      ew_button_attribs(W);
      goto done;
   case EWbuttonCANCEL:
   case GFXEV_QUIT:
      goto done;
   case EWbuttonICONCHOOSE:
      ic= choose_icon();
      if (!ic) break;
      gfx_set_text(EWbuttonICON,ic);
      break;
   }
done:
  gfx_leave_screen();
}

void ew_button(Uwidget_t *W)
{
  ew_button_generic(W, "Button Attributes");
}

void ew_toggle(Uwidget_t *W)
{
  ew_button_generic(W,"Toggle Button Attributes");
}

static void ew_textonly(Uwidget_t *W,char *title)
{
  gfx_SCRewSIMPLE();
  gfx_set_text(EWsimpleTITLE,title);
  ewf_write();
  ewf_text(EWsimpleTEXT,W,"text");
  if (ewloop(EWsimpleOK,EWsimpleCANCEL))
  {
    ewf_read();
    ewf_text(EWsimpleTEXT,W,"text");
  }
  gfx_leave_screen();
  
}

void ew_textentry(Uwidget_t *W)
{
  ew_textonly(W,"Text Entry Attributes");
}

void ew_checkbox(Uwidget_t *W)
{
  ew_textonly(W,"Check Box Attributes");
}

static void ew_scrollbar_attribs(Uwidget_t *W)
{
  ewf_enum(EWscbVERTICAL, W, "vertical", "0@Horizontal#1@Vertical");
}

void ew_scrollbar(Uwidget_t *W)
{
  gfx_SCRewSCROLL();
  ewf_write();
  ew_scrollbar_attribs(W);
  if (ewloop(EWscbOK,EWscbCANCEL))
  {
    ewf_read();
    ew_scrollbar_attribs(W);
  }
  gfx_leave_screen();
}

void ew_none(Uwidget_t *W)
{
  gfx_ask("Information","!OK","Widget %s of class %s\n"
                              "has  no editable attributes", W->ident,
    W->klass->ident);
}

void ew_radio_attribs(Uwidget_t *W)
{
  ewf_text(EWradioTEXT, W,"text");
  ewf_text(EWradioGROUP,W,"group");
}

void ew_radio(Uwidget_t *W)
{
  gfx_SCRewRADIO();
  ewf_write();
  ew_radio_attribs(W);
  if (ewloop(EWradioOK,EWradioCANCEL))
  {
    ewf_read();
    ew_radio_attribs(W);
  }
  gfx_leave_screen();
}

static void ew_slider_attribs(Uwidget_t *W)
{
  ewf_enum(EWsliderORIENT, W,"orientation", "0@Horizontal#1@Vertical");
}

void ew_slider(Uwidget_t *W)
{
 gfx_SCRewSLIDER();
 ewf_write();
 ew_slider_attribs(W);
 if (ewloop(EWsliderOK,EWsliderCANCEL))
 {
   ewf_read();
   ew_slider_attribs(W);
 } 
 gfx_leave_screen();
}


static void ew_iconlist_attribs(Uwidget_t *W)
{
  ewf_text(WewICONLISTfg,W,"fg");
  ewf_text(WewICONLISTbg,W,"bg");
  ewf_text(WewICONLISThpad,W,"hpad");
  ewf_text(WewICONLISTvpad,W,"vpad");
  ewf_text(WewICONLISTmaxtextw,W,"max_text_width");
  ewf_text(WewICONLISTmulti,W,"multi_select");
}

void ew_iconlist(Uwidget_t *W)
{
 gfx_SCRewICONLIST();
 ewf_write();
 ew_iconlist_attribs(W);
 if (ewloop(WewICONLISTok,WewICONLISTcancel))
 {
   ewf_read();
   ew_iconlist_attribs(W);
 } 
 gfx_leave_screen();
}

