#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include "expr.h"
#include "widget.h"
#include "strbuf.h"
#include "eval.h"
#include "library.h"
#include "ptrbuf.h"
#include "gen.h"
#include "n-eval.h"
#include "bytecode.h"

static int first_widget_no, last_widget_no;


static void make_widget_layout(Uscreen_t *scr)
{
  int n_locals;
  bc_buffer_t *bc;

  if (scr->wl_code) bc_buffer_free(scr->wl_code);
  scr->wl_code= NULL;
  scr->wl_nlocals= 0;

   // FIXME: get some errors in here.
  bc= compile_widget_layout(scr, &n_locals);
  if (!bc)
  {
    printf("compilation failed\n");
    return ;
  }
  scr->wl_code= bc;
  scr->wl_nlocals= n_locals;
}

static void gen_widget_names(FILE *f)
{
  Uscreen_t *scr;
  Uwidget_t *W;
  Uscreen_t **SS; int N;
  int i;
  int last;

  SS= get_all_screens(&N);
     // Warning!
     // These are from gfx.h if you change those, you should
     // come here and change these also.
#ifndef GFXEV_LAST
#define GFXEV_LAST 14
#endif
#ifndef GFXRESERVED
#define GFXRESERVED 24
#endif

  fprintf(f," /* WIDGET NAMES BELOW */\n");
  fprintf(f,"enum {\n ");
  char buf[2000];
  buf[0]= 0;

  last= GFXEV_LAST+GFXRESERVED+1;
  sprintf(buf,"=%d", last);
  first_widget_no= last;
  for(i=0;i<N;i++)
  {
   scr= SS[i];
   for(W=scr->widgets;W;W=W->next)
   {
     fprintf(f,"  %s%s,\n", W->ident,buf);
     W->global_number= last;
     if (!strcmp(W->klass->ident,"PixGrid"))
     {
       sprintf(buf,"=%s+GFXEV_LAST",W->ident);
       last+= GFXEV_LAST;
     }
     else
     {
       last++;
       buf[0]= 0;
     }
   }
  }
  fprintf(f,"};\n");
  free(SS);
  last_widget_no= last-1;
}

typedef struct { int len; FILE *file; } array_out_t;

static void AOs(array_out_t *AO, char *str)
{
  int L;
  L= strlen(str);
  if (AO->len+L>72) { fprintf(AO->file,"\n"); AO->len= 0; }
  AO->len+= L;
  fprintf(AO->file,"%s",str);
}

static void AOinit(array_out_t *AO, FILE *f, char *init)
{
  AO->file= f;
  AO->len= 0;
  AOs(AO,init);
}

static void AOi(array_out_t *AO, int V)
{
  char buf[20];
  sprintf(buf,"%d,",V);
  AOs(AO,buf);
}

static void AOc(array_out_t *AO, unsigned char V)
{
  char buf[20];
  sprintf(buf,"%d,",V);
  AOs(AO,buf);
}
static void AOcc(array_out_t *AO, unsigned char *D, int len)
{
  int i;
  for(i=0;i<len;i++) AOc(AO, D[i]);
}

static void AObc(array_out_t *AO, bc_buffer_t *bc)
{
  int i;
  for(i=0;i<bc->len;i++) AOc(AO, bc->data[i]);
}

static void declare_widget_list(Uwidget_t **WL, int n, FILE *f)
{
  int i;
  array_out_t A;
  AOinit(&A, f, " static int SWID[]= { ");
  for(i=0;i<n;i++) AOi(&A,WL[i]->global_number);
  fprintf(f,"};\n");
}

static Uwidget_t** get_widget_list_breadthfirst(Uscreen_t *scr,int *Rn)
{
  // It's important to get the list in breadthfirst fashion,
  // this is because we want to calculate the size&pos for the
  // parent widgets first, and then the child widgets. 
  ptrbuf_t *queue;
  ptrbuf_t *result;
  Uwidget_t *W;
  int i;
  size_t Nww; Uwidget_t **ww;
  int nresult; // #widgets in result so far

  result= ptrbuf_new(20); // holds the widgets
  queue= ptrbuf_new(20); // holds the screens
  nresult= 0;

  ptrbuf_put(queue, scr);
  while((scr=ptrbuf_get(queue)))
  {
    scr->windex_start= nresult;
    for(W=scr->widgets;W;W=W->next)
    {
      ptrbuf_put(result, W);
      nresult++;
      scr->windex_end= nresult;
      for(i=0;i<W->n_subs;i++) ptrbuf_put(queue, W->subs[i]);
    }
  }
  ptrbuf_free(queue,NULL,NULL);
  ptrbuf_free(result, (void***) &ww, &Nww);
  *Rn= Nww;
  return ww;
}

static void Gcreate_widget
  (Uscreen_t *scr,Uwidget_t *W, bc_buffer_t *bc)
{
  if (W->klass->construct)
  {
     (*W->klass->construct)(scr,W,bc);
  }
  else 
  {
     fprintf(stderr,"unhandled widget class : %s\n" , W->klass->ident);
  }
}

void bc_buffer_put_param(bc_buffer_t *bc,int param)
{
  bc_buffer_put_int(bc, param);
  bc_buffer_put(bc, GUII_IPARAM);
}


/***********************************************************************
   The screen object. This contains everything to create a screen.
   gui_screen_start parses and uses this object.

   Format:

   n_widgets
   n_direct_widgets
   widget_number_list  : an array of integers, each integer is a widget ID.

   // screen layout code
   n_locals
   code_length
   code

   // constructor code
   code_length
   code

   // widget layout code
   n_locals
   code_length
   code

   n_subs   : number of subscreens due to panels
     <for each subscreen>
   parent_index :   index of the containing widget within widget_number_list
   start_index  :   index of the first child widget within widget_number_list
   end_index    : 1+index of the last child widget within widget_number_list
   n_locals
   code_length
   code

   The output is a byte array. This byte array may be put in a file
   directly, or be emitted as C code, depending on the situation.
 ***********************************************************************/

static void objo_screen_header
  (bc_buffer_t *B,int n_widgets,int n_direct_widgets,Uwidget_t **widgets)
{
  int i;
  bc_buffer_put_param(B, n_widgets);
  bc_buffer_put_param(B, n_direct_widgets);

  for(i=0;i<n_widgets;i++) bc_buffer_put_int(B, widgets[i]->global_number);
  bc_buffer_put_int(B, n_widgets);
  bc_buffer_put(B, GUII_INTARRAY);
  bc_buffer_put(B, GUII_PPARAM);
}

static void objo_screen_layout(bc_buffer_t *B, Uscreen_t *scr)
{
  bc_buffer_put_param(B, scr->sl_nlocals);
  bc_buffer_put_param(B, scr->sl_code->len);
  bc_buffer_append(B, scr->sl_code);
}

static void objo_constructor
   (bc_buffer_t *B, Uscreen_t *scr, int n_widgets, Uwidget_t **widgets)
{
  int i;
  bc_buffer_t *cons;

  cons= bc_buffer_new(1000);
  for(i=0;i<n_widgets;i++) Gcreate_widget(scr, widgets[i], cons);

  bc_buffer_put_param(B, cons->len);
  bc_buffer_append(B, cons);
  bc_buffer_free(cons);
}

static void objo_widget_layout(bc_buffer_t *B, Uscreen_t *scr)
{
  bc_buffer_put_param(B, scr->wl_nlocals);
  bc_buffer_put_param(B, scr->wl_code->len);
  bc_buffer_append(B, scr->wl_code);
}

static void objo_sub_layout(bc_buffer_t *B, int pindex, Uscreen_t *S)
{
  bc_buffer_put_param(B, pindex);
  bc_buffer_put_param(B, S->windex_start);
  bc_buffer_put_param(B, S->windex_end);
  objo_widget_layout(B, S);
}

static void gen_screen_object(Uscreen_t *scr, unsigned char **RO, int *RL)
{
  Uwidget_t *W;
  int i,j;
  Uscreen_t *S;
  int n_subs;
  bc_buffer_t *B;
  Uwidget_t **ww; int Nww; // this is the list of ALL widgets,
                           // including those in the sub widgets.
  int Nw;
    // This is the number of widgets directly contained in
    // the screen. This number of widgets appear first in
    // the widgetlist, since it's a breadth-first list.
    // Nww and Nw are used differently. All widgets are
    // subjected to some operations such as translating
    // them by the origin of the screen. For widgets directly
    // contained in the screen, we do other things such as
    // showing them. The difference between these two stem
    // from panel widgets, which can contain other widgets.

  B= bc_buffer_new(10*1024);

  ww= get_widget_list_breadthfirst(scr,&Nww);

  for(Nw=0,W=scr->widgets;W;W=W->next,Nw++)
    if (W!=ww[Nw]) break;

  for(i=0,n_subs=0;i<Nww;i++)
    if (ww[i]->n_subs) n_subs+= ww[i]->n_subs;

  objo_screen_header(B, Nww, Nw, ww);
  objo_screen_layout(B, scr);
  objo_constructor(B, scr, Nww, ww);
  objo_widget_layout(B, scr);
  
  bc_buffer_put_param(B, n_subs);

  // Now we need to layout the subscreens. We do that using the
  // breadth-first widget list, which gives us a breadth-first
  // sub-screen list. 

  for(i=0;i<Nww;i++)
  {
    W= ww[i];
    if (!W->n_subs) continue;
    for(j=0;j<W->n_subs;j++)
    {
      S= W->subs[j];
      objo_sub_layout(B, i, S);
    }
  }

  *RO= B->data;
  *RL= B->len; 
  free(B);
}

// A screen layout object is a simple thing. it's just
//   n_locals
//   code_length
//   code
static void make_screen_layout(Uscreen_t *scr)
{
  bc_buffer_t *code;
  int nlocals;
 
  code= compile_screen_layout(scr, &nlocals);
  scr->sl_nlocals= nlocals;
  scr->sl_code= code;
}

static void prt_array_pretty(FILE *f,unsigned char *data, int len)
{
  int x;
  int i,w;
  unsigned char K;
  x= 0;
  fprintf(f,"   ");
  for(i=0;i<len;i++)
  {
    K= data[i];
    w= K>99 ? 4: (K>9 ? 3: 2);
    if (x+w>72) { x= 0; fprintf(f,"\n   "); }
    fprintf(f,"%d,", K);
    x+= w;
  }
}

static void gen_screen_func(Uscreen_t *scr,FILE *f,int mode)
{
  unsigned char *data;
  int len;
  fprintf(f,"void gfx_%s()",scr->ident);
  if (mode=='H') { fprintf(f,";\n"); return ; }

  gen_screen_object(scr, &data, &len);

  fprintf(f,"{\n");
  fprintf(f,"  static int code_len= %d;\n", len);
  fprintf(f,"  static unsigned char code[]= {\n");
  prt_array_pretty(f, data, len);
  fprintf(f,"  };\n");
  fprintf(f,"  gui_screen_start(code);\n");
  fprintf(f,"}\n");

  free(data);
}


static char *replace_suffix(char *str,char *sfx)
{
  int i;
  str= strdup(str);
  for(i=strlen(str)-1;i>0;i--)
     if (str[i]=='.') { strcpy(str+i,sfx); break; }
  return  str;
}


static void gen_colorimg_resource(FILE *f,resource_t *R)
{
  int i;
  int L;
  fprintf(f,"RES_COLORIMG(%s,\"User resource\")",R->ident);
  L= R->length/4;
  for(i=0;i<L;i++)
  {
    if (i%5==0) fprintf(f,"\n");
    fprintf(f,"0x%08x,", *(uint32_t*) (R->data+4*i));
  }
  fprintf(f,"\nRES_END\n");
}

static void gen_resources(FILE *f)
{
  resource_t *R;
  for(R=resources;R;R=R->next)
  {
   switch(R->type)
   {
   case TYP_COLORIMG:
    gen_colorimg_resource(f,R);
    break;
   }
  }
}

static void gen_resource_iface(FILE *f)
{
  resource_t *R;
  for(R=resources;R;R=R->next)
  {
   switch(R->type)
   {
   case TYP_COLORIMG:
    fprintf(f, "extern unsigned int %s[];\n", R->ident);
    break;
   }
  }
}

static void generate_header(FILE *out)
{
  const char *Chdr;
  int i,N;
  Uscreen_t **SS,*S;
  Uiconset_t *IS;

  print_code("gfx.h", out);

  gen_widget_names(out);
  SS= get_all_screens(&N);

  fprintf(out,"// Screen numbers\n");
  fprintf(out,"enum {\n");
  
  for(i=0;i<N;i++)
  {
    S= SS[i];
    if (S->type) continue;
    S->screen_number= i;
    fprintf(out, "  GFXS_%s,\n", S->ident);
  }
  fprintf(out,"};\n");
  free(SS);

  fprintf(out,"enum gfx_icons {\n");
  for(i=0;i<iconsets.len;i++)
  {
     IS= iconsets.data[i];
     if (i) fprintf(out, "  %s,\n", IS->name);
       else fprintf(out, "  %s=1,\n", IS->name);
  }
  fprintf(out,"};\n");


  gen_resource_iface(out);
}

const char *print_to_mark(const char *code, const char *mark, FILE *f)
{
  const char *z;
  z= strstr(code,mark);
  if (z)
    { while(*z!=0 && *z!='\n') z++;
      if (*z=='\n') z++; }
  else
    z= code+strlen(code);
  if (f)
    fwrite(code, 1, z-code,f);
  return z;
}

/***********************************************************************
  A module consists of the following:
    1- screen data
    2- resources (icons etc)
    3- fonts

  The format is :
     <icon data>
     <screen data>

  Here is the icon data format:
    n_icons
    <for each icon>
      n_images
      <for each image>
        min_dpi
        data_length
        <data> : bytes 

  Here is the format for screen data.
    first_widget_no
    last_widget_no
    n_screens
      <for each screen>
    size
    data

  Look at gfx_data in the library code.
 ***********************************************************************/

static void print_int(FILE *out, int K)
{
  fprintf(out, "%d,%d,%d,%d,\n", (K>>24)&0xff, (K>>16)&0xff, 
               (K>>8)&0xff, K&0xff);
}

static void gen_iconimage(FILE *out, Uicon_t *icon)
{
  unsigned char x;
  int V, L,i;
  fprintf(out, "// ICON \n");
  print_int(out, icon->dpi);
  print_int(out, icon->data_length);
  fprintf(out,"\n // ICON DATA\n");
  for(i=0;i<icon->data_length;i++)
  {
    x= icon->data[i];
    V= x/100 ? 4 : ( x/10 ? 3 : 2 );
    if (L+V>=76) { fprintf(out,"\n"); L= 0; }
    L+= V;
    fprintf(out,"%d,",x);
  }
}

static void gen_icon(FILE *out, Uiconset_t *IS)
{
  int i;
  print_int(out, IS->n_icons);
  for(i=0;i<IS->n_icons;i++)
     gen_iconimage(out, IS->icons+i);
}

static void gen_icons(FILE *out)
{
  int i;
  print_int(out, iconsets.len);
  for(i=0;i<iconsets.len;i++)
    gen_icon(out, iconsets.data[i]);
}

static void gen_module(int nSS, Uscreen_t **SS, FILE *out)
{
  int i;
  int n_screens;
  unsigned char *data; int len;

  n_screens= 0;
  // count the number of top level screens
  for(i=0;i<nSS;i++)
   if (!SS[i]->type) n_screens++;

  fprintf(out,"const unsigned char mdata[]= {\n");
  gen_icons(out);

  print_int(out, first_widget_no);
  print_int(out, last_widget_no);
  print_int(out, n_screens);

  for(i=0;i<nSS;i++)
  {
    if (SS[i]->type) continue; // skip sub-screens.
    gen_screen_object(SS[i], &data, &len);
    fprintf(out, "// screen %s size= %d\n", SS[i]->ident, len);
    print_int(out, len);
    fprintf(out,"\n");
    prt_array_pretty(out, data,len);
    free(data);
  }
  fprintf(out,"};\n");

  fprintf(out,"void gfx_localdata() { gfx_data(mdata); }\n");
}

static void generate_source_file(char *fn)
{
  FILE *out;
  const char *Ccode;
  int i,N;
  Uscreen_t **SS,*S;

  out= fopen(fn, "w");
  if (!out) return ; // FIXME: give error message

  fprintf(out,"void gfx_data(const unsigned char *data);\n");

  SS= get_all_screens(&N);

  // compile all widget & screen layout code
  for(i=0;i<N;i++)
  {
     make_widget_layout(SS[i]);
     if (!SS[i]->type) make_screen_layout(SS[i]);
  }

  gen_module(N, SS, out);
 
  free(SS);
  fclose(out);
}

void generate_code(char *hfilenm,char *cfilenm)
{
  if (hfilenm)
  {
    FILE *out;
    out= fopen(hfilenm,"w");
    if (out) { generate_header(out); fclose(out); }
  }
  if (cfilenm) generate_source_file(cfilenm);
}

