/***********************************************************************
 *                                                                     *
 *                                                                     *
 ***********************************************************************
 This is the main file for the code generator.
 ***********************************************************************/
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdint.h>
#include "ptrbuf.h"
#include "expr.h"
#include "widget.h"
#include "save.h"
#include "eval.h"
#include "gen.h"
#include "path.h"


xpp_t *xpp;

resource_t *resources;

static char *cfile,*hfile;
static char *inputf;
static int load_ok;


static void init_expression_parser()
{
  xpp= xpp_new();

  xpp_binary_operator(xpp, "+", EXP_ADD, 50, 'L');
  xpp_binary_operator(xpp, "-", EXP_SUB, 50, 'L');
  xpp_binary_operator(xpp, "*", EXP_MUL, 60, 'L');
  xpp_binary_operator(xpp, "/", EXP_DIV, 60, 'L');
  xpp_binary_operator(xpp, "%", EXP_MOD, 60, 'L');
  xpp_prefix_operator(xpp, "-", EXP_NEG);
  xpp_prefix_operator(xpp, "+", EXP_PLUS);
  xpp_parenthesis(xpp, "(", ")", EXP_PAREN_R);
  xpp_binary_operator(xpp, ".", EXP_DOT, 200, 'L');
}

static int bool_arg(int *argc,char **argv,char *flag, int *Rv)
{
  int i;
  if (*argc<2 || strcmp(argv[1],flag)) return 0;
  for(i=1;i<*argc-1;i++)
    argv[i]= argv[i+1];
  *Rv= 1;
  (*argc)--;
  return 1;
}

static int str_arg(int *argc,char **argv,char *flag, char **Rv)
{
  int i;
  if (*argc<3 || strcmp(argv[1],flag)) return 0;
  *Rv= argv[2];
  for(i=1;i<*argc-2;i++)
    argv[i]= argv[i+2];
  (*argc)-= 2;
  return 1;
}

static char *abs_or_relainp(char *val)
{
  if (!val) return NULL;
  if (val[0]=='/') return strdup(val);
  if (inputf) 
      return path_absolute(path_child(path_dir(inputf), val));
  return path_absolute(val);
}

static int load_initial_file(char *prgname,int force)
{
  load_ok= 0;
  prj_settings_t settings;

  if (inputf)
  {
    Uscreen_t *S; resource_t *R;
    if (load_file(inputf, &S,&R,&settings)) 
    {
      resources= R;
      screens= screen= S;
      load_ok= 1;
      if (!hfile) hfile= abs_or_relainp(settings.hdr_file);
      if (!cfile) cfile= abs_or_relainp(settings.src_file);
    }
  }

  if (!load_ok)
  {
    if (force)
    {
      fprintf(stderr,"%s: an input file must be specified\n", prgname);
      return 1;
    }
    screen= screens= create_screen();
  }
  return 0;
}

static void usage(char **argv)
{
   exit(fprintf(stderr,"usage: %s [-h header] [-c source] inputfile\n",
                     argv[0]));
}


int main(int argc,char **argv)
{
  init_expression_parser();

  inputf= cfile= hfile= NULL;

  while(argc>1)
  {
    if (str_arg(&argc,argv,"-h", &hfile) ||
        str_arg(&argc,argv,"-c", &cfile))
      continue;
    if (argv[1][0]=='-')
      usage(argv);
    inputf= argv[1];
    argv[1]= argv[0];
    argv++; argc--;
  }

  if (!inputf) usage(argv);

  if (load_initial_file(argv[0],1))
     return 1;
  if (!cfile) cfile= path_child(path_dir(inputf),"gfx.c");
  if (!hfile) hfile= path_child(path_dir(inputf),"gfx.h");
  fprintf(stderr,"generating code\n  header: %s\n  source: %s\n",hfile,cfile);
  generate_code(hfile,cfile);
  return 0;
}

