#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include "expr.h"
#include "widget.h"
#include "gfx.h"

#define LODEPNG_PRIVATE
#include "../libsrc/components/031-library-png.c"



static Uiconset_t *sel_iconset; static int sel_iconset_idx= -1;
static int sel_icon_idx= -1;

static void put_icon_in_list(int idx, Uicon_t *IC)
{
  char buf[200];
  sprintf(buf,"%dx%d    %d DPI", IC->w, IC->h, IC->dpi);
  gfx_insert(WicIMAGES, idx, buf, NULL, NULL);
}

static int print_icon_attribs(int idx)
{
  // FIXME: do the preview
  if (!sel_iconset || idx<0 || idx>=sel_iconset->n_icons) 
  {
     gfx_set_text(WicDPI, "");
     gfx_set_text(WicIMAGEpath,"");
     return -1;
  }
  else
  {
     char buf[50];
     sprintf(buf,"%d",sel_iconset->icons[idx].dpi);
     gfx_set_text(WicDPI, buf);
     gfx_set_text(WicIMAGEpath, sel_iconset->icons[idx].orig_file_name);
     return idx;
  }
}

static void select_icon(int idx)
{
  idx= print_icon_attribs(idx);
  gfx_set_selected(WicIMAGES, idx);
  sel_icon_idx= idx;
  if (idx==-1) return ;
}

static void select_iconset(int idx, Uiconset_t *IS)
{
  int i;
  sel_iconset_idx= idx;
  if (IS==sel_iconset) return ;
  sel_iconset= IS;
  gfx_pause();
  gfx_clear(WicIMAGES);
  sel_icon_idx= -1;
  if (IS) 
    for(i=0;i<IS->n_icons;i++)
    {
      put_icon_in_list(i, IS->icons+i);
    }
  gfx_set_text(WicIMAGEpath, "");
  gfx_set_text(WicDPI, "");
  gfx_set_selected(WicIMAGES,-1);
  gfx_resume();
}

static void load_icon()
{
  
}

static void show_iconlist()
{
  int i;
  Uiconset_t *IC;
  gfx_pause();
  for(i=0;i<iconsets.len;i++)
  {
    IC= iconsets.data[i];
    gfx_insert(WicICONS, -1, IC->name, IC, NULL);
  }
  gfx_resume();
}

static char *strip_spaces(char *K)
{
  char *R; int i,j;
  R= malloc(strlen(K)+1);
  for(i=0,j=0;K[i];i++)
    if (K[i]>' ' && K[i]!=127) R[j++]= K[i];
  R[j]= 0;
  return R;
}

static int cclass(char K)
{
  if (K>='a' && K<='z') return 'a';
  if (K>='A' && K<='Z') return 'a';
  if (K=='_') return 'a';
  if (K>='0' || K<='9') return '9';
  return '.';
}

static int is_valid_identifier(char *R)
{
  int i,K;
  if (cclass(R[0])!='a') return 0;
  for(i=1;R[i];i++) 
  {
    K= cclass(R[i]);
    if (K!='a' && K!='9') return 0;
  }
  return 1;
}

static char *get_icon_name()
{
  gfx_event_t ev;
  int q,i;
  char *R,*V;
  Uiconset_t *IS;
 
  q= 0;
  R= NULL;
  gfx_SCRicASKname();

  while(!q)
  {
   if (!gfx_event(NULL,&ev)) continue;
   switch(ev.type)
   {
   case GFXEV_QUIT: 
   case WicanCANCEL: q= 1; break;
   case WicanOK:
     V= gfx_get_text(WicanTEXT);
     if (!V) break;
     R= strip_spaces(V);
     free(V);
     if (!is_valid_identifier(R)) 
      { gfx_set_text(WicanSTATUS,"Please enter a valid C identifier.");
        free(R); R= NULL;
        break; }
     for(i=0;i<iconsets.len;i++)
     {
       IS= iconsets.data[i];
       if (!strcmp(IS->name,R)) break;
     }
     if (i!=iconsets.len)
     {
       free(R); R= NULL;
       gfx_set_text(WicanSTATUS,"Entered name is not unique.");
       break;
     }
     q= 1;
     break;
   }
  }
  gfx_leave_screen(); 
  return R;
}


static void add_icon()
{
  char *name;
  Uiconset_t *IS;
  name= get_icon_name();
  if (!name) return ;
  IS= par_alloc(&iconsets, sizeof(*IS));
  IS->name= name;
  gfx_insert(WicICONS, -1, IS->name, IS, NULL);
}

static void free_icon(Uicon_t *IC)
{
  free(IC->data);
}

static void remove_icon()
{
  Uiconset_t *IS;
  int i;
  if (sel_iconset_idx==-1) return ;
  iconsets.data[sel_iconset_idx]= iconsets.data[iconsets.len-1];
  iconsets.len--;
  IS= sel_iconset;
  gfx_remove(WicICONS, sel_iconset_idx, NULL);
  select_iconset(-1, NULL);
  for(i=0;i<IS->n_icons;i++)
   free_icon(IS->icons+i);
  free(IS->icons);
  free(IS);
}

static int ask_dpi()
{
  gfx_event_t ev;
  int q;
  int dpi;
  q= 0;
  dpi= 0;
  gfx_SCRicad();
  gfx_focus(WicadTEXT);
  while(!q)
  {
    if (!gfx_event(NULL,&ev)) continue;
    switch(ev.type)
    {
    case GFXEV_QUIT: q= 1; break;
    case WicadOK: dpi= atoi(gfx_get_text(WicadTEXT)); q= 1; break;
    }
  }
  gfx_leave_screen();
  return dpi;
}

static int read_file(unsigned char **data, int *datalen, char *fn)
{
  FILE *f;
  int L;
  unsigned char *D;
  f= fopen(fn,"rb");
  if (!f) return 1;
  fseek(f, 0, SEEK_END);
  L= ftell(f);
  if (L<=0) { fclose(f); return 1; }
  D= malloc(L);
  fseek(f, 0, SEEK_SET);
  fread(D, 1, L, f);
  fclose(f);
  *data= D;
  *datalen= L;
  return 0;
}

static void add_image()
{
  static char *file_types[]= { "PNG Images (*.png)", NULL };
  Uicon_t *icons;
  Uiconset_t *IS;
  char *fn;
  int dpi;
  int R;
  Uicon_t IC;
  int i,j,pos;
  unsigned char *Odata; unsigned int Ow, Oh;

  if (!sel_iconset) return ;
  fn= gfx_select_file(0, NULL, file_types);
  if (!fn) return ;
  dpi= ask_dpi();

  R= lodepng_decode32_file(&Odata, &Ow, &Oh, fn);
  if (R)
  {
    gfx_ask("PNG Decoding Error", "!OK", "File %s:\n\n%s\n\n",fn,  
             lodepng_error_text(R));
    return ;
  }
  free(Odata);

  IC.orig_file_name= fn;
  IC.w= Ow;
  IC.h= Oh;
  IC.dpi= dpi;

  if (read_file(&IC.data, &IC.data_length,fn))
     return ;

  icons= malloc(sizeof(icons[0])*(IS->n_icons+1));
  IS= sel_iconset;
  
  for(i=0,j=0;i<IS->n_icons;i++,j++)
    if (IS->icons[i].dpi>dpi) break;
    else icons[j]= IS->icons[i];

  pos= j;

  icons[j++]= IC;
  for(;i<IS->n_icons;i++,j++)
    icons[j]= IS->icons[i];

  if (IS->icons) free(IS->icons);
  IS->icons= icons;
  IS->n_icons++;

  put_icon_in_list(pos, &IC);
  select_icon(pos);
}

void icon_screen()
{
  gfx_event_t ev;
  gfx_SCRicons();
  int q;
 
  show_iconlist();
  q= 0;

  while(!q)
  {
   if (!gfx_event(NULL,&ev)) continue;
   switch(ev.type)
   {
   case WicICONS: select_iconset(ev.key,ev.usr); break;
   case WicIMAGES: select_icon(ev.key); break;
   case WicADDicon: add_icon(); break;
   case WicDELicon: remove_icon(); break;
   case WicADDimage: add_image(); break;
   case WicUPDATE:
   case WicDELimage:
     break;
   case GFXEV_QUIT:
   case WicOK: q=1; break;
   }
  }
  gfx_leave_screen();

  sel_iconset_idx= -1;
  sel_icon_idx= -1;
  sel_iconset= NULL;
  // WicIMAGEpath,
  // WicDPI,                      
   // case WicPIX,
}

char *choose_icon()
{
  char *R;
  int q;
  int i;
  Uiconset_t *IS;
  gfx_event_t ev;
again:
  R= NULL;
  q= 0;
  gfx_SCRiconselect();
  for(i=0;i<iconsets.len;i++)
  {
    IS= iconsets.data[i];
    gfx_insert(WicseLIST, -1, IS->name, IS, NULL);
  }
  while(!q)
   if (gfx_event(NULL,&ev)) switch(ev.type)
   {
   case GFXEV_QUIT:
   case WicseCANCEL: R= NULL; q=1; break;
   case WicseOK: if (!R) break; q= 1; break;
   case WicseLIST: IS= ev.usr; R= IS->name; break;
   case WicseMANAGE: gfx_leave_screen(); icon_screen(); goto again;     
   }
  gfx_leave_screen();
  if (R) R= strdup(R);
  return R;
}
