#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdint.h>
#include "expr.h"
#include "ptrbuf.h"
#include "widget.h"

par_t iconsets;

Uscreen_t *screen,*screens;

Uwidget_t* make_widget(Uklass_t *klass,int x1,int y1,int x2,int y2)
{
  Uwidget_t *W,*P;
  char buf[200];
  static int LW;
  W= calloc(1,sizeof(*W));
  sprintf(buf,"W%d",LW++);
  W->ident= strdup(buf);
  W->klass= klass;
  if (!strcmp(klass->ident,"ScrollBar"))
  {
    W->pos.S[wWID]= strdup("gui_scrollbar_width");
    W->pos.S[wHEI]= strdup("gui_scrollbar_height");
    W->pos.E[wWID]= xpp_run(xpp,W->pos.S[wWID]);
    W->pos.E[wHEI]= xpp_run(xpp,W->pos.S[wHEI]);
  }
  W->pos.Dx= x1;
  W->pos.Dy= y1;
  W->pos.Dw= x2-x1;
  W->pos.Dh= y2-y1;

  if (!screen->widgets)
    screen->widgets= W;
  else
  {
    for(P=screen->widgets;P->next;P=P->next) ;
    P->next= W;
  }

  return W;
}

void free_widget(Uwidget_t *w)
{
  // FIXME: there are a lot more things here
  free(w);
}

Uwidget_t *find_widget_bydpos(int x,int y)
{
  Uwidget_t *W;
  for(W=screen->widgets;W;W=W->next)
    if (W->pos.Dx<=x && W->pos.Dy<=y && (W->pos.Dx+W->pos.Dw)>x
        && (W->pos.Dy+W->pos.Dh)>y) return W;
  return NULL;
}

Uwidget_t *find_widget_in_screen(Uscreen_t *S,char *id)
{
  Uwidget_t *W,*R;
  int i;
  if (!S) return NULL;
  for(W=S->widgets;W;W=W->next)
  {
    if (!strcmp(W->ident,id)) return W;
    for(i=0;i<W->n_subs;i++)
      if ((R=find_widget_in_screen(W->subs[i],id)))
        return R;
  }
  return NULL;
}

Uwidget_t *find_widget(char *id)
{
  Uwidget_t *W;
  Uscreen_t *S;
  for(S=screens;S;S=S->next)
    if ((W= find_widget_in_screen(S, id)))
      return W;
  return NULL;
}

Uscreen_t *create_screen()
{
  char buf[200];
  Uscreen_t *S;
  static int last_scr;
  sprintf(buf,"SCREEN_%d",last_scr++);
  S=calloc(1,sizeof(*S));
  S->ident= strdup(buf);
  return S;
}

static void get_sub_screens(ptrbuf_t *B, Uscreen_t *S)
{
  Uwidget_t *W;
  int i;

  ptrbuf_put(B, S);
  for(W=S->widgets;W;W=W->next)
   for(i=0;i<W->n_subs;i++)
     get_sub_screens(B, W->subs[i]);
}

Uscreen_t** get_all_screens(int *Rn)
{
  ptrbuf_t *B;
  Uscreen_t *S,**SS;
  size_t n;

  B= ptrbuf_new(20);
  for(S=screens;S;S=S->next)
   get_sub_screens(B,S);
  ptrbuf_free(B, (void***) &SS, &n);
  *Rn= n;
  return SS;
}

void free_var(Uvar_t *v)
{
 // FIXME: do this properly
  free(v);
}

void free_upos(Upos_t *p)
{
  // FIXME: do this properly
  // do NOT free(v), it's part of the
  // containing struct.
}

void destroy_screen(Uscreen_t *S)
{
  Uwidget_t *w;
  Uvar_t *v;

  free(S->ident);
  while(S->widgets)
  {
    w= S->widgets;
    S->widgets= w->next;
    free_widget(w);
  }

  while(S->vars)
  {
    v= S->vars;
    S->vars= v->next;
    free_var(v);
  }

  free_upos(&S->pos);
  free(S);
}

int get_icon_number(char *name)
{
  int i;
  Uiconset_t *IS;
  if (!name) return 0;
  for(i=0;i<iconsets.len;i++)
  {
    IS= iconsets.data[i];
    if (!strcmp(IS->name,name)) return i+1;
  }
  return 0;
}
