
enum { TYP_INT, TYP_STR, TYP_COLORIMG, TYP_LIST };

enum { wTOP, wBOT, wLEF, wRIG, wWID, wHEI };


enum { EXP_ADD=XPP_LAST, EXP_SUB, EXP_DIV, EXP_MOD, EXP_MUL,
       EXP_NEG, EXP_PLUS, EXP_PAREN_R, EXP_DOT };


typedef struct bcbuffer
{
  int size, len;
  unsigned char *data;
  struct bcbuffer *left, *right;
} bc_buffer_t;

typedef struct widget Uwidget_t;
typedef struct screen Uscreen_t;

typedef struct
{
  expr_t *E[6];
  char *S[6];
  int Dx,Dy,Dw,Dh; // display x,y,w,h

  int Db,Dr;      // display bottom and right, used only for
                  // evaluation, never stored or loaded
} Upos_t;

typedef struct field
{
  char *name;
  int type;
  void *data;
  struct field *next;
} field_t;

typedef struct
{
  char *ident;
  int nfields;
  field_t *fields; // these fields contain the default values, if any
  void (*create)(Uwidget_t *,FILE*);
  void (*edit)(Uwidget_t *);
  void (*construct)(Uscreen_t *,Uwidget_t *,bc_buffer_t*);
} Uklass_t;


struct widget 
{
  Uklass_t *klass;
  char *ident;
  Upos_t pos;
  field_t *fields; // these contain the user-set values
  struct widget *next;
  void *priv;

  int n_subs;
  Uscreen_t **subs;

  int mark; unsigned int mark_color;

  int global_number;
     // this is the numerical value of the widget's 
     // identifier. we need to determine this ourselves
     // and not rely on the enum because we need to 
     // output these as numbers where no compiler is
     // available
};

enum { VAR_NONE, VAR_X, VAR_Y, VAR_W, VAR_H };

typedef struct uvar
{
  char *ident;
  char *str;
  expr_t *expr;
  int vartype;
  struct uvar *next;

  int sto;
} Uvar_t;


struct screen
{
  char *ident;
  Uwidget_t *widgets;
  Uvar_t *vars;
  int type; // 0= normal 1= subscreen
  Upos_t pos; // screen position & size
              // should only use WIDTH and HEIGHT..

  /* These are very curious structures. Now, in the library code,
     there is a field called widgets[] in the screen structure.
     Let the library screen be called S, and this subscreen
     be called U. 
        S->widgets[windex_start]= number( firstwidget(U) );
        S->widgets[windex_end]  = number( lastwidget(U) ) + 1;
     So, all widgets of U are recorded between the offsets
     [windex_start,windex_end) within S->widgets. This lets me
     calculate the sub-screen layouts. This is complicated stuff, beh. */
  int windex_start, windex_end;

  struct screen *next;

  /* Now, getting more complicated. Layout is now done by bytecode. 
     The following hold the information regarding the 'widget' layout.
     This is buffered until we can output the layout object for the
     screen. */
  bc_buffer_t *wl_code;
  int wl_nlocals;

  /* The same thing as above, but for the screen size and position. */
  bc_buffer_t *sl_code;
  int sl_nlocals;

  // This is the index of the screen within the list of
  // screens in the module.
  int screen_number;
};

typedef struct
{
  char *target;
  expr_t *expr;
  char *str;
  int source; // 'V' or 'W'
  void *ptr;  // corresponding variable or widget
  int subsource; // in case it's a widget, which coordinate (wTOP etc)

  int *store;
} unknown_t;

typedef struct resource
{
  struct resource *next;
  char *ident;
  int length;
  int type;
  unsigned char *data;
} resource_t;

typedef struct
{
  int canvas_w, canvas_h;
  char *hdr_file, *src_file;
} prj_settings_t;

extern Uklass_t *klasses[];
extern Uscreen_t *screen,*screens;
extern xpp_t *xpp;
extern resource_t *resources;


Uwidget_t* make_widget(Uklass_t *klass,int x1,int y1,int x2,int y2);
void free_widget(Uwidget_t *w);
Uwidget_t *find_widget_bydpos(int x,int y);
Uwidget_t *find_widget_in_screen(Uscreen_t *S,char *id);
Uwidget_t *find_widget(char *id);
Uscreen_t *create_screen();
Uscreen_t** get_all_screens(int *Rn);

Uklass_t *choose_klass();

void destroy_screen(Uscreen_t *S);

// Widget Editors
void ew_button    (Uwidget_t *);
void ew_canvas    (Uwidget_t *);
void ew_checkbox  (Uwidget_t *);
void ew_iconlist  (Uwidget_t *);
void ew_label     (Uwidget_t *);
void ew_panel     (Uwidget_t *);
void ew_progressbar(Uwidget_t *);
void ew_radio     (Uwidget_t *);
void ew_scrollbar (Uwidget_t *);
void ew_slider    (Uwidget_t *);
void ew_textentry (Uwidget_t *);
void ew_toggle    (Uwidget_t *);
void ew_none      (Uwidget_t *);
void ew_filelist  (Uwidget_t *);


// Warning!
/// these are from gfx.c. I need them because I must output these
// numerically on occasion such as when I'm constructing buttons etc.
//
enum {
  GFXCLASS_LABEL, 
  GFXCLASS_LOGVIEW,
  GFXCLASS_SPINNER,
  GFXCLASS_PROGRESSBAR, 
  GFXCLASS_RADIO,   
  GFXCLASS_PANEL,
  GFXCLASS_BUTTON,
  GFXCLASS_TOGGLE,
  GFXCLASS_PIXGRID,
  GFXCLASS_CHECKBOX,
  GFXCLASS_DIRLIST,
  GFXCLASS_FILELIST,
  GFXCLASS_TEXTENTRY,
  GFXCLASS_SCROLLBAR,
  GFXCLASS_LIST,
  GFXCLASS_ICONLIST,
  GFXCLASS_SLIDER,
  GFXCLASS_DROPDOWN,
  GFXCLASS_CANVAS,
};

typedef struct
{
  int size, len;
  void **data;
} par_t;

void* par_alloc(par_t *par, size_t size);
void par_add(par_t *par, void *ptr);
void par_reset(par_t *par);


typedef struct
{
  int dpi;
  int w, h;
  unsigned char *data;
  int data_length;
  char *orig_file_name;
} Uicon_t;

typedef struct Uiconset
{
  char *name;
  int n_icons;
  Uicon_t *icons; 
} Uiconset_t;

extern par_t iconsets;


int get_icon_number(char *iname);

char *choose_icon();

