
/** BEGIN: GFX HEADER **/

typedef struct
      {
  int   type;
  int   x, y;
  int   modifiers;
  int   button;
  int   key;
  char *string;
  void *usr;
  int wide;
  double flo0, flo1;
  unsigned long time;
      }
    gfx_event_t;

enum
{
  GFXEV_BUTTON_PRESS=1, GFXEV_BUTTON_RELEASE, GFXEV_DOUBLE_CLICK,
  GFXEV_KEY_PRESS,      GFXEV_KEY_RELEASE,
  GFXEV_POINTER_MOVE,   
  GFXEV_READY_READ, GFXEV_READY_WRITE, GFXEV_READY_EXCEPT,
  GFXEV_QUIT, GFXEV_JAVARESULT, GFXEV_TIMEOUT,
  GFXEV_FOCUS,
  GFXEV_LAST,
};

#define GFXEV(X,Y) (((X)<<8)|(GFXEV_##Y)|(1<<31))

#define GFXRESERVED 24

/* The event routine returns an event from the queue. Normally, the return
value is non-zero. In case of a failure, return value is zero. FIXME:
this will be fixed so that a 'failure' event will be returned.  */

int gfx_event(unsigned int *__unused_msecs,gfx_event_t *Revent);

enum
{ 
   GFX_K_PRT=1, 
   GFX_K_BACKSPACE,     GFX_K_TAB,      GFX_K_RETURN, GFX_K_ESCAPE,
   GFX_K_INSERT,        GFX_K_DELETE,   GFX_K_HOME,   GFX_K_END,
   GFX_K_PAGEUP,        GFX_K_PAGEDOWN,
   GFX_K_LEFT,          GFX_K_RIGHT,    GFX_K_DOWN,   GFX_K_UP,
   GFX_K_CONTROL=0x100, GFX_K_SHIFT= 0x200 , GFX_K_ALT= 0x400,
};

#define GFX_VK_NOP -1

typedef struct { char *str[3]; int span; int func; }       gfx_vkeydef_t;
typedef struct { gfx_vkeydef_t keys[15][50]; }             gfx_vkbdef_t;

#ifndef GFX_WIN
int gfx_init(char *name, int width, int height, char *title);
#endif

#ifdef GFX_WIN
void gfx_init(HINSTANCE instance, int w, int h, char *window_title);
#endif

void gfx_close();

int gfx_event(unsigned int *msecs,gfx_event_t *Revent);
void gfx_timer_start(int ident, int ms);
void gfx_timer_stop(int ident);

void gfx_geometry(unsigned int *W, unsigned int *H);
void gfx_set_sysbar_height(int hh);
int gfx_get_dpi();

// These two pause and resume drawing on the screen.
// Put your updates between these two for better performance.
void gfx_pause();
void gfx_resume();

void gfx_monitor_file(int fd, int ident, int R,int W,int X);
void gfx_stop_monitor_file(int fd,int ident);

#ifdef GFX_ANDROID
#ifdef JNI_H_
JavaVM *gfx_get_javavm();
#endif
#ifdef ANDROID_NATIVE_ACTIVITY_H
ANativeActivity* gfx_get_activity();
#endif
extern void (*gfx_app_get_state)(unsigned char **Rdata,int *Rlen);
extern void (*gfx_app_stop)();
void gfx_soft_keyboard(int show);
#endif

void gfx_leave_screen();

void gfx_focus(int wno);
void gfx_text_data(int font, const char *str, 
                   int *Rw, int *Rh,int *Ra,unsigned char **Rd);


char* gfx_get_text(int wno);
void gfx_set_text(int wno,char *str);

void gfx_clear(int wno);
void gfx_insert(int wno,int pos, char *text,void *user,unsigned int* icon);
void gfx_remove(int wno, int pos, void **Ruser);

void gfx_set_fgbg(int wno, int fg, int bg);
void gfx_set_selected(int wno,int sel);
int  gfx_get_selected(int wno);

void gfx_set_state(int wno, int state);
int  gfx_get_state(int wno);

// #W.SCROLLBAR
void gfx_scrollbar_set(int wno, double start, double end, double step);

int gfx_ask(char *title, char *butstr, char *fmt, ...);
char* gfx_select_file(int mode,char *init,char **types);

// W.ICONLIST
void gfx_iconlist_updates(int wno, int upd);
void gfx_iconlist_area(int wno, int *Rw, int *Rh);
void gfx_iconlist_select(int wno, int item, int sel);
void gfx_iconlist_permute(int wno, int N, int *P);
void gfx_iconlist_clear_selection(int wno);
void gfx_iconlist_get_selection(int wno,int *Rn, int **Rs);
void gfx_iconlist_insert
      (int wno, int pos, char *text, 
       unsigned int w, unsigned int h, unsigned int *data,
       int ival, void *pval);

// W.DIRLIST
void gfx_dirlist_set_directory(int wno,char *dir);
char *gfx_dirlist_get_directory(int wno);
void gfx_dirlist_refresh(int wno);

// W.FILELIST
void gfx_filelist_set_directory(int wno,char *dir);
void gfx_filelist_set_types(int wno, char **types);
char** gfx_filelist_get_files(int wno);
char *gfx_filelist_get_directory(int wno);
void gfx_filelist_refresh(int wno);
char *gfx_filelist_get_selected(int wno);

// #W.LOGVIEW
void gfx_logview_clear(int wno);
void gfx_set_log(int wno);
void loprintv(const char *fmt, va_list ap);
void loprintf(const char *fmt, ...);
void gfx_logview_printf(int wno, const char *fmt, ...);
void gfx_logview_write(int wno, const char *txt, size_t len);

// W.PIXGRID
void gfx_pixgrid_attribs
   (int pg, int *Rshown, unsigned int *Rw, unsigned int *Rh);
int gfx_pixgrid_buffer
  (int pg, int *Rbpl, int *Rbpp, int *Rorder, unsigned char **Rbuffer);
void gfx_pixgrid_update(int pg, int x,int y,int w, int h);
void gfx_pixgrid_putsimple(int pg, unsigned int *img);
void gfx_pixgrid_put
    (int pg, int x,int y,
     unsigned int *img, unsigned int iW,unsigned int iH,
     int src_x, int src_y, 
     int copy_w, int copy_h);

// W.PROGRESS
void gfx_progressbar_total(int wno, int total);
void gfx_progressbar_set(int wno, int value);

// #W.SPINNER
// Integer
  void gfx_spinner_set_imin(int wno, int min);
  void gfx_spinner_set_imax(int wno, int max);
  void gfx_spinner_set_iquant(int wno, int qua);
  void gfx_spinner_set_ivalue(int wno, int value);
   int gfx_spinner_get_ivalue(int wno);
// Float
  void gfx_spinner_set_fmin(int wno, double min);
  void gfx_spinner_set_fmax(int wno, double max);
  void gfx_spinner_set_fquant(int wno, double qua);
  void gfx_spinner_set_fvalue(int wno, double value);
double gfx_spinner_get_fvalue(int wno);

// #W.SLIDER
double gfx_slider_get(int wno);
void gfx_slider_set(int wno, double pos);

// #D.VIRTKEYB
void gfx_virtual_keyboard(int num, gfx_vkbdef_t *D);
void gfx_hide_virtual_keyboard();
void gfx_edit_field(int widget, int kbnum);

// #W.CANVAS
void gfx_canvas_setsize(int wno, int new_w, int new_h);
int gfx_canvas_get_buffer
   (int wno, unsigned int *W, unsigned int *H, unsigned int **B);
void gfx_canvas_put
       (int wno, int dx,int dy, int w, int h,
        unsigned int *source, int sw, int sh, int sx, int sy);
void gfx_canvas_put_graymap
       (int wno, unsigned int color, int dx,int dy, int w, int h,
        unsigned char *source, int sw, int sh, int sx, int sy);
void gfx_canvas_update(int wno, int x, int y, int w, int h);

void gfx_canvas_print(int wno, int font, int color, int x, int y, char *str);
void gfx_canvas_draw_rectangle(int wno,int color, int x,int y,int w,int h);
void gfx_canvas_fill_rectangle(int wno,int color, int x,int y,int w,int h);

int gfx_sprite_new(int wno,int func, int w,int h, uint32_t *data);
void gfx_sprite_show(int wno,int sp,int sho);
void gfx_sprite_move(int wno,int sp, int x,int y);
void gfx_sprite_remove(int wno, int sp);

void gfx_canvas_set_zoom_params
  (int wno, int zf, int grid, unsigned int gridcolor,int w,int h);
void gfx_canvas_zoom(int wno,int x,int y);
void gfx_canvas_hide_zoom(int wno);

void gfx_canvas_begin_report_mouse_position(int wno);
void gfx_canvas_end_report_mouse_position(int wno);


void gfx_canvas_recsel_begin(int wno, gfx_event_t *ev);
void gfx_canvas_recsel_end(int wno,int *Rx1, int *Ry1, int *Rx2, int *Ry2);

/** END: GFX HEADER **/
