
#define NFONTS 3

static fontdef_t fontdefs[NFONTS]= {
{ fontdata0, 16 },
{ fontdata1, 16 },
{ fontdata1, 24 },
};

static font_t fonts[NFONTS];

/*****************
  Scale has to be computed in a very careful way. A font may contain a
  lot of characters from different languages, all of which may not be 
  of our concern. For example, if a font contains mathematical symbols,
  the tall integral sign affects the whole font. And if you specify that
  all glyphs should fit in a given height, the system will fit the
  integral sign to that height and normal characters will be much smaller.
  A better way to find the scale is to look at some typical characters.
  For example, the descent of jpgyq can be combined with ascent of 
  M1KTF. The total of these two values can be scaled to fit within a
  certain height. 
 *****************/

static void compute_chr_vmetrics
   (stbtt_fontinfo *info, int chr, int *rA, int *rD)
{
  int x0, y0, x1, y1;
  int A, D;
  x0= y0= x1= y1= 0; // shut up the compiler about possible uninitval
  stbtt_GetCodepointBox(info, chr, &x0,&y0,&x1,&y1);
  if (y0<0) A= -y0; else A= 0;
  if (y1<0) D=   0; else D= y1;
  if (rA) *rA= A;
  if (rD) *rD= D;
}

static void compute_str_vmetrics
  (stbtt_fontinfo *info, char *str, int *RmaxA, int *RmaxD)
{
  int maxA, maxD, A,D;
  maxA= maxD= 0;
  while(*str)
  {
    compute_chr_vmetrics(info, str[0], &A, &D);
    maxA= imax2(maxA, A);
    maxD= imax2(maxD, D);
    str++;
  }
  if (RmaxA) *RmaxA= maxA;
  if (RmaxD) *RmaxD= maxD;
}

static float compute_font_scale(stbtt_fontinfo *info,int pix)
{
  int A,D,H;
  compute_str_vmetrics(info, "ygqPMK1TFQpj",&A,&D);
  H= A + D;
  return  pix / (float) H;
}

static void fonts_init()
{
  int i;
  if (fonts[0].cache) return ;
  for(i=0;i<NFONTS;i++)
  {
    fonts[i].cache= womap_new();
    stbtt_InitFont(&fonts[i].finfo, fontdefs[i].data,
                   stbtt_GetFontOffsetForIndex(fontdefs[i].data,0));
    fonts[i].height= fontdefs[i].size * getdpi()/86;
    fonts[i].scale= compute_font_scale(&fonts[i].finfo,  fonts[i].height);
  }
  static int* htargets[]= { &font1height, &font2height, &font3height,
                            &font4height, &font5height, NULL };
  for(i=0;htargets[i];i++)
    if (i>=NFONTS) htargets[i][0]= fonts[0].height;
             else  htargets[i][0]= fonts[i].height;
}

/***********************************************************
  Implementing correct rendering in the general case is a 
  really complicated issue. All I will do is to ignore the
  case where the glyph extend outside its own box. For example,
  italics do this. I will implement something better later.
  For now, I need to do this as soon as possible. 

  I will render each character as individually, and then simply
  copy the whole thing into the combined buffer. The buffer
  will also hold the necessary spaces. 
 ***********************************************************/

static fchr_t* font_render_char(font_t *F, int chr)
{
  fchr_t *CC;
  int gindex; 
  int bw, bh,x,y,xoff,yoff,advance,lbearing, w,h;
  unsigned char *bmap, *big;
  CC= womap_find(F->cache, chr);
  if (CC) return CC;
  gindex= stbtt_FindGlyphIndex(&F->finfo, chr);
  if (!gindex) gindex= stbtt_FindGlyphIndex(&F->finfo, 'A');
  if (!gindex) gindex= 1; // I don't know man..
  xoff= yoff= w= h= 0; // shut the compiler up
  bmap= stbtt_GetGlyphBitmap(&F->finfo,0,F->scale, gindex, &w,&h, &xoff, &yoff);
  stbtt_GetGlyphHMetrics(&F->finfo,gindex,&advance, &lbearing);

  bw= 0;
  if (xoff<0) { bw= -xoff; xoff= 0;}

  bw= imax2(bw,imax2(xoff+w,(int) (advance*F->scale)));
  bh= h;
  big= calloc(bw*bh,sizeof(big[0]));
  for(y=0;y<h;y++)
    for(x=0;x<w;x++)
      big[(x+xoff)+y*bw]= bmap[x+y*w];
  stbtt_FreeBitmap(bmap,NULL);
  
  CC= malloc(sizeof(*CC));
  CC->chr= chr;
  CC->width= bw; CC->height= bh;
  CC->data= big;
  CC->ascent= -yoff;
  CC->descent= h+yoff;
  womap_insert(F->cache, chr, CC);
  return CC;
}


static void font_render
        (font_t *F, uint32_t *str, size_t L,
        int *RW, int *RH, int *RA,
        unsigned char **Rdata)
{                                       
  fchr_t **chars;
  int i, x, y, maxA, maxD, w,h, cw, ch, ca, xx, yy;
  unsigned char *data, *cd;

  fonts_init();

  chars= alloca(L*sizeof(chars[0]));
  for(i=0;i<L;i++)
    chars[i]= font_render_char(F, str[i]);

  maxA= maxD= 0;
  w= 0;
  for(i=0;i<L;i++)
  {
    maxA= imax2(maxA, chars[i]->ascent);
    maxD= imax2(maxD, chars[i]->descent);
    w+= chars[i]->width;
  }
  h= maxA+maxD;

  if (RW) *RW= w;
  if (RH) *RH= h;
  if (RA) *RA= maxA;
  if (!Rdata) return ;

  *Rdata= data= calloc(w*h,sizeof(data[0]));
  xx= 0; yy= maxA;
  for(i=0;i<L;i++)
  {
    cw= chars[i]->width; ch= chars[i]->height; 
    ca= chars[i]->ascent; cd= chars[i]->data;
    for(y=0;y<ch;y++)
       for(x=0;x<cw;x++)
          data[ (xx+x) + (yy-ca+y)*w ]= cd[x+y*cw];
    xx+= cw;
  }
}

static void* __unused_freetype_lib[]= {
(void*) stbtt_BakeFontBitmap, (void*) stbtt_GetBakedQuad, (void*) stbtt_GetScaledFontVMetrics,
(void*) stbtt_PackBegin, (void*) stbtt_PackEnd , (void*) stbtt_PackFontRange,
(void*) stbtt_PackSetOversampling, (void*) stbtt_PackSetSkipMissingCodepoints,
(void*) stbtt_GetPackedQuad, (void*) stbtt_GetNumberOfFonts, (void*) stbtt_GetFontVMetricsOS2,
(void*) stbtt_GetFontBoundingBox, (void*) stbtt_GetCodepointHMetrics,
(void*) stbtt_GetCodepointKernAdvance, (void*) stbtt_IsGlyphEmpty,
(void*) stbtt_GetCodepointShape, (void*) stbtt_FreeShape, (void*) stbtt_GetCodepointBitmap,
(void*) stbtt_MakeCodepointBitmap, (void*) stbtt_MakeCodepointBitmapSubpixelPrefilter,
(void*) stbtt_GetCodepointBitmapBox, (void*) stbtt_FreeSDF, (void*) stbtt_GetCodepointSDF,
(void*) stbtt_FindMatchingFont, (void*) stbtt_CompareUTF8toUTF16_bigendian,
(void*) stbtt_GetFontNameString,
};

