
/***********************************************************************
 *                                                                     *
 *                             User Timers                             *
 *                                                                     *
 ***********************************************************************
  Now it's time for me to implement user defined timers. These will be
  used instead of the timeout argument to the gfx_event. Finally.

  Each user timer will be a once-shot timer, which will fire once and
  stop. The timer is then kept on the list until it's removed by the
  user. Later on, the timer can be restarted. This will let me have
  static timer identifiers..
 ***********************************************************************/

typedef struct { int enabled, ident, left; } usr_timer_t;

static usr_timer_t *usrtimers;
static int susrtimers, nusrtimers;

void gfx_timer_start(int ident, int ms)
{
  int i;
  usr_timer_t *T;
  if (ident<0) return ;
  T= NULL;
  if (!usrtimers)
  {
   usrtimers= calloc(sizeof(usrtimers),10);
   susrtimers= 10;
  }
  for(i=0;i<nusrtimers;i++)
  {
    T= usrtimers+i;
    if (T->ident==ident) break;
  }
  if (i==nusrtimers)
  {
    T= array_make1space(nusrtimers, (void**) &usrtimers, nusrtimers, 
                          &susrtimers, sizeof(usrtimers[0]));
    nusrtimers++;
  }
  T->ident= ident;
  T->enabled= 1;
  T->left= ms;
}

void gfx_timer_stop(int ident)
{
  int i;
  usr_timer_t *T;
  if (ident<0) return ;
  for(i=0;i<nusrtimers;i++)
  {
    T= usrtimers+i;
    if (T->ident==ident) { T->enabled= 0; break; }
  }
}

/***********************************************************************
  In order to activate these things, I should probably make some sort
  of priority queue, but I'm not very keen on that. I'll just go thru
  the list linearly. 
 ***********************************************************************/
static int have_user_timer(int *Rms)
{
  int i;
  int R= 0;
  int ms= 0;
  for(i=0;i<nusrtimers;i++)
   if (usrtimers[i].enabled)
   {
     if (usrtimers[i].left<0) usrtimers[i].left= 0;
     if (R) ms= imin2(ms, usrtimers[i].left);
       else ms= usrtimers[i].left;
     R= 1;
   }
  *Rms= ms;
  return R;
}

/***********************************************************************
  When some time has elapsed, this function is called. It creates an
  event if there is a timed out timer.
 ***********************************************************************/
static int elapsed_user_timer(int ms,int retev, gfx_event_t *event)
{
  int i;
  int R;
  R= 0;
  for(i=0;i<nusrtimers;i++)
  {
    if (!usrtimers[i].enabled) continue;
    usrtimers[i].left -= ms;
    if (usrtimers[i].left<0) usrtimers[i].left= 0;
    if (usrtimers[i].left==0 && retev)
    {
      event->type= GFXEV(usrtimers[i].ident,TIMEOUT);
      usrtimers[i].enabled= 0;
      R= 1;
      retev= 0;
    }
  }
  return R;
}



