
/***********************************************************************
 *                                                                     *
 *                           File Monitoring                           *
 *                                                                     *
 ***********************************************************************
  The graphics system should also be able to select() descriptors other
  than the X connection or the mouse device etc. For this purpose,
  I will introduce file monitors.

  For Android, there already is a polling mechanism called a looper. On
  this platform, we register the required events with this struct and
  then dispatch them directly from get_event() etc. without consulting
  the table again.
 ***********************************************************************/

typedef struct
{
  int fd;
  int ident;
  int events; // 1=read 2=write 4=except
} file_monitor_t;

static file_monitor_t fmon[100]; // safe limit, I guess
static int n_fmon;

void gfx_monitor_file(int fd, int ident, int R,int W,int X)
{
  file_monitor_t *F;
  if (n_fmon==100) return ;
  if (!R && !W && !X) return ;
  F= fmon + n_fmon;
  n_fmon++;
  F->fd= fd;
  F->ident= ident;
  F->events= (R?1:0) | (W?2:0) | (X?4:0);
#ifdef GFX_ANDROID
  ALooper_addFd(my_app->looper, fd, ident | (1<<16), 
         (R ? ALOOPER_EVENT_INPUT:0) |
         (W ? ALOOPER_EVENT_OUTPUT:0) |
         (X ? ALOOPER_EVENT_ERROR|ALOOPER_EVENT_HANGUP
                                 |ALOOPER_EVENT_INVALID:0), NULL, my_app);
#endif
}

void gfx_stop_monitor_file(int fd,int ident)
{
  int i;
  for(i=0;i<n_fmon;i++)
    if ( (fd==-1 || fmon[i].fd==fd) &&
         (ident==-1 || fmon[i].ident==ident))
    {
#ifdef GFX_ANDROID
      ALooper_removeFd(my_app->looper, fmon[i].fd);
#endif
      fmon[i]= fmon[n_fmon-1];
      n_fmon--;
      i--;
    }
}


static void populate_select_set
  (int *Rmaxfd, fd_set *rfd, fd_set *wfd, fd_set *xfd)
{
  int maxfd;
  int i;
  maxfd= *Rmaxfd;
  for(i=0;i<n_fmon;i++)
  {
    if (fmon[i].events&1) FD_SET(fmon[i].fd, rfd);
    if (fmon[i].events&2) FD_SET(fmon[i].fd, wfd);
    if (fmon[i].events&4) FD_SET(fmon[i].fd, xfd);
    maxfd= imax2(maxfd, fmon[i].fd);
  }
  *Rmaxfd= maxfd;
}

static int get_file_monitor_event
  (fd_set *rfd, fd_set *wfd, fd_set *xfd, gfx_event_t *Revent)
{
  static int last_pos;
  file_monitor_t *F;
  int i;

  if (!n_fmon) return 0;
  if (last_pos>=n_fmon) last_pos%=n_fmon;
  for(i=0;i<n_fmon;i++)
  {
     F= fmon + (i+last_pos)%n_fmon;
     if ((F->events&4) && FD_ISSET(F->fd, xfd)) 
          { Revent->type= GFXEV_READY_EXCEPT; goto found; }
     if ((F->events&2) && FD_ISSET(F->fd, wfd)) 
          { Revent->type= GFXEV_READY_WRITE; goto found; }
     if ((F->events&1) && FD_ISSET(F->fd, rfd)) 
          { Revent->type= GFXEV_READY_READ; goto found; }
  }
  return 0;
found:
  Revent->key= F->ident;
  Revent->wide= F->fd;
  last_pos= (last_pos+1)%n_fmon;
  return 1;
}


