
/***********************************************************************
 *                                                                     *
 *                         X11 Event Handling                          *
 *                                                                     *
 ************************************************************************/

static void tv2msecs(struct timeval *tv, unsigned int* msecs)
{
  if (!msecs) return ;
  *msecs= tv->tv_sec*1000 + tv->tv_usec/1000;
}

static int get_event(gfx_event_t *Revent)
{
  int x,R,maxfd;
  XEvent event;
  fd_set rfd, xfd, wfd;
  XExposeEvent *expose;

  struct timeval tv,*tvp;
  unsigned int delay,remaining,elapsed;
  unsigned int user_delay, sys_delay;
  int have_sys_delay, have_user_delay;

  user_delay= sys_delay= 0; // warnUNINVAR

  XSync(x_display,False);
again:

  // This is done before checking for timers because I want to
  // give X11 a chance to send an event in case the user is asking
  // for a really rapid timer. 
  if (XPending(x_display)) goto have_event; 

  have_sys_delay= gui_timer_get(&sys_delay); 

  if (have_sys_delay && sys_delay==0 && gui_timer_elapsed(0,Revent))
    return 1;

  have_user_delay= have_user_timer(&user_delay);
  if (have_user_delay && user_delay==0 && elapsed_user_timer(0,1,Revent))
    return 1;

  tvp= &tv;
  switch(flag2(have_user_delay,have_sys_delay))
  {
  case 0: tvp= NULL; break;
  case 1: delay= sys_delay; break;
  case 2: delay= user_delay; break;
  case 3: delay= imin2(user_delay,sys_delay); break;
  }

  if (tvp)
  {
    tv.tv_sec= delay/1000;
    tv.tv_usec= (delay%1000)*1000;
  }   

  x= ConnectionNumber(x_display);

  FD_ZERO(&rfd); FD_ZERO(&xfd); FD_ZERO(&wfd);
  FD_SET(x, &rfd); FD_SET(x, &xfd);
  maxfd= x;

  populate_select_set(&maxfd, &rfd, &wfd, &xfd);

  R= select(maxfd+1, &rfd, &wfd, &xfd, tvp);

  if (R<0)
  {
    if (errno==EINTR)
    {
       usleep(1000);
       goto again;
    }
    return -1; // FIXME: log this stuff.
  }
  if (tvp)
  {
    tv2msecs(&tv, &remaining);
    elapsed= delay-remaining;
    switch(flag2(have_user_delay,have_sys_delay))
    {
    case 1: if (gui_timer_elapsed(elapsed,Revent)) return 1; break;
    case 2: if (elapsed_user_timer(elapsed,1,Revent)) return 1; break;
    case 3: 
        elapsed_user_timer(elapsed, 0, Revent);
        if (gui_timer_elapsed(elapsed,Revent)) return 1;
        if (elapsed_user_timer(0, 1, Revent)) return 1;
        break;
    }
  }

  if (get_file_monitor_event(&rfd,&wfd,&xfd,Revent)) return 1;

  // Well, there was an exception on the X11 fd. Most probably
  // the connection to the X server was lost or there is a memory
  // thrashing problem. Let's just return and have the program
  // fail in some other way.
  if (FD_ISSET(x, &xfd)) return 0; // FIXME: log this stuff.

  // All that remains is the X11 file descriptor. If that is not
  // set as well, there is something wrong here. Let's go back
  // and start all over again.
  if (!FD_ISSET(x, &rfd)) goto again;

have_event:
  XNextEvent(x_display, &event);

  switch(x_keyboard_event(x_keyboard,&event,Revent))
  {
  case 0: break;
  case 1: return 1;
  case -1: goto again;
  }

  switch(event.type)
  {
   case Expose:
     expose= &event.xexpose;
                                                  // FIXME: This zero should be
                                                  // something else
     XCopyArea(x_display, x_pixmap, x_window, DefaultGC(x_display,0), 
               expose->x, expose->y, expose->width, expose->height,
               expose->x, expose->y);
     goto again; 
   case MotionNotify:
     // if (gfx_recsel_event(&event)) goto again;
     Revent->type= GFXEV_POINTER_MOVE; 
     Revent->x= event.xmotion.x; 
     Revent->y= event.xmotion.y; 
     Revent->modifiers= event.xmotion.state; 
     break;
   case ButtonPress:
     Revent->type= GFXEV_BUTTON_PRESS; 
     Revent->x= event.xbutton.x; 
     Revent->y= event.xbutton.y; 
     Revent->modifiers= event.xbutton.state; 
     Revent->button= event.xbutton.button;
     Revent->time= event.xbutton.time;
     break;
   case ButtonRelease:
     Revent->type= GFXEV_BUTTON_RELEASE; 
     Revent->x= event.xbutton.x; 
     Revent->y= event.xbutton.y; 
     Revent->modifiers= event.xbutton.state; 
     Revent->button= event.xbutton.button;
     Revent->time= event.xbutton.time;
     break;
   case ClientMessage: 
     if (event.xclient.message_type!=wm_protocols) goto again;
     if (event.xclient.data.l[0]!=wm_delete_window) goto again;
     Revent->type= GFXEV_QUIT;
     break;
   default:
//     printf("unknown event type: %d\n", event.type);
     goto again;
  }

  return 1; 
}


